/*
                           Device Record Management

	Functions:

	int DevRecIsReallyMounted(char *dev)
	int DevRecIsAllocated(int dev_num)

	int DevRecDoMount(int dev_num)
	int DevRecDoUnmount(int dev_num)

	int DevRecRefreshList()
	int DevRecAllocate()
	void DevRecDelete(int rec_num)
	void DevRecDeleteAllRecords()

	void DevRecDoComfUnmountAll()

 */

#include "endeavour.h"



/* *********************************************************************
 *
 */
int DevRecIsReallyMounted(char *dev)
{
#if defined (__linux__)

	FILE *fp;
	int i;
	int strc;
	char **strv;
	char *line;
	int status = 0;


	if(dev == NULL)
	    return(0);


	fp = fopen("/proc/mounts", "r");
	if(fp == NULL)
	    return(0);

	line = NULL;
	while(1)
	{
	    free(line); line = NULL;
	    line = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
	    if(line == NULL)
		break;

	    /* Explode line. */
	    strv = strexp(line, &strc);

	    /* Check if givin device name matches dev string. */
	    if(strc >= 1)
	    {
		if(strv[0] != NULL)
	        {
		    if(!strcmp(strv[0], dev))
		    {
			status = 1;

                        /* Free allocated exploded strings. */
                        for(i = 0; i < strc; i++)
                        {
                            free(strv[i]); 
                            strv[i] = NULL;
                        }
                        free(strv); strv = NULL;
                        strc = 0;

			break;
		    }
		}
	    }

	    /* Free allocated exploded strings. */
	    for(i = 0; i < strc; i++)
	    {
		free(strv[i]);
		strv[i] = NULL;
	    }
	    free(strv); strv = NULL;
	    strc = 0;
	}


        free(line); line = NULL;

	fclose(fp);
	fp = NULL;

	return(status);

#else

	/* Return false for all other platforms. */
	return(0);

#endif
}



int DevRecIsAllocated(int dev_num)
{
	if((devrec == NULL) ||
           (dev_num < 0) ||
	   (dev_num >= (int)total_devrecs)
	)
	{
	    return(0);
	}
	else if(devrec[dev_num] == NULL)
        {
            return(0);
        }
	else
        {
            return(1);
        }
}



/*
 *	Performs a mount using the information from the given device
 *	record.
 *
 *	Current working directory will be changed.
 */
int DevRecDoMount(int dev_num)
{
        int retries;
	char stringa[512 + PATH_MAX + NAME_MAX];


	/* Error checks. */
        if(!DevRecIsAllocated(dev_num))
            return(-1);
	if(devrec[dev_num]->cmd_mount == NULL)
	    return(-2);


	/* Remove program output spool file. */
	SFRemoveProgStdout();


	/* Change working directory. */
        if(chdir("/") == 0)
	{
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
	}


	/* Issue mount command. */
	strncpy(stringa, devrec[dev_num]->cmd_mount, 512 + PATH_MAX + NAME_MAX);
	stringa[512 + PATH_MAX + NAME_MAX - 1] = '\0';
	ExecBAOE(stringa, fname.prog_output, fname.prog_output);


        /* Wait untill mount is successful. */
        for(retries = 0; retries < option.mount_timeout; retries++)
        {
            if(DevRecIsReallyMounted(devrec[dev_num]->devname))
            {
                /* Update mount state. */
                devrec[dev_num]->is_mounted = 1;
                break;
            }
            else
            {
                sleep(1);
            } 
        }

	/* Print warning if failed to mount. */
	if(!devrec[dev_num]->is_mounted)
	{
	    sprintf(stringa,
"Unable to mount:\n\n\
     Device: %s\n\
         To: %s\n\n\
The device failed to mount or has timmed out.   Check\n\
all configuration files to ensure that the correct values are\n\
givin for mounting this device.   In addition, make sure the\n\
device has power and (if it is a removeable device) ensure\n\
that it is plugged in and that the media is inserted.\n",
		devrec[dev_num]->devname,
		devrec[dev_num]->path
	    );
	    printdw(&err_dw, stringa);
	}
	else
	{
	    /* Else print any warnings from mount command. */
	    SFShowOutputOnDialog(&err_dw);
	}


	/* Refresh device records list. */
	DevRecRefreshList();


	return(0);
}



/*
 *	Unmounts a device from the information from the given
 *	device record.
 *
 *	Current working directory will be changed.
 */
int DevRecDoUnmount(int dev_num)
{
	int retries;
        char stringa[512 + PATH_MAX + NAME_MAX];


	/* Error checks. */
	if(!DevRecIsAllocated(dev_num))
	    return(-1);
        if(devrec[dev_num]->cmd_unmount == NULL)
            return(-2);

        /* Check if device is unmountable. */
        if(devrec[dev_num]->flags & DEV_TF_NOUNMOUNT)
            return(-3);


        /* Remove program output spool file. */
        SFRemoveProgStdout();


        /* Change working directory. */
        if(chdir("/") == 0)
        {
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
        }


        /* Issue unmount command. */
        strncpy(stringa, devrec[dev_num]->cmd_unmount, 512 + PATH_MAX + NAME_MAX);
        stringa[512 + PATH_MAX + NAME_MAX - 1] = '\0';
        ExecBAOE(stringa, fname.prog_output, fname.prog_output);


        /* Wait untill unmount is successful. */
	for(retries = 0; retries < option.mount_timeout; retries++)
	{
            if(DevRecIsReallyMounted(devrec[dev_num]->devname))
            {
		sleep(1);
            }
            else
            {
                /* Update mount state. */
                devrec[dev_num]->is_mounted = 0;

		break;
	    }
        }

        /* Print warning. */
        if(devrec[dev_num]->is_mounted)
        {
            sprintf(stringa,
"Unable to unmount:\n\n\
     Device: %s\n\
         At: %s\n\n\
The device failed to unmount or has timmed out.   Check if\n\
any other programs are currently using this device and that\n\
their current working directories do not referance a child path\n\
of this device's mounted destination.\n",
                devrec[dev_num]->devname,
                devrec[dev_num]->path      
            );
            printdw(&err_dw, stringa);
        }
        else
        {
            /* Else print any warnings from mount command. */
            SFShowOutputOnDialog(&err_dw);
        }


        /* Refresh device records list. */
        DevRecRefreshList();


	return(0);
}



/* ******************************************************************
 *
 *	Refreshes mounted devices, checking if they are really mounted.
 *
 *	Also refetches device capacity and other info from the
 *	device find command.
 */
int DevRecRefreshList()
{
	int i;
	char tmp_cmd[CMD_MAX];
	FILE *fp;
	char *linebuf;

	char **strv;
	int strc;

	char *strptr;

	char dev[PATH_MAX + NAME_MAX];
	off_t capacity_total;
	off_t capacity_used;
	off_t capacity_available;
	double capacity_percent;
	char path[PATH_MAX + NAME_MAX];


	/* ************************************************************** */
	/* Update all loaded device records' mount states. */
        for(i = 0; i < (int)total_devrecs; i++)
        {
            if(!DevRecIsAllocated(i))
                continue;

            if(DevRecIsReallyMounted(devrec[i]->devname))
                devrec[i]->is_mounted = 1;
            else
                devrec[i]->is_mounted = 0;
        }


	/* Update all mounted device records' capacity information. */
	if(command.df != NULL)
	{
	    /* Clear program stdout output file. */
	    SFRemoveProgStdout();

	    /* Format and run device find command. */
	    strncpy(tmp_cmd, command.df, CMD_MAX);
	    tmp_cmd[CMD_MAX - 1] = '\0';
	    /* Execute in background and append to stdout and stderr. */
            ExecBAOE(tmp_cmd, fname.prog_output, fname.prog_output);


	    /* Open and read output file. */
	    fp = fopen(fname.prog_output, "rb");
	    if(fp != NULL)
	    {
		/* Skip first line (heading). */
		linebuf = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
		free(linebuf); linebuf = NULL;

		/* Begin reading each line after first (heading) line. */
		linebuf = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
		while(linebuf != NULL)
		{
		    /* Strip '\n' and '\r' from linebuf, replace with '\0'. */
		    while((strptr = strchr(linebuf, '\n')) != NULL)
			*strptr = '\0';
                    while((strptr = strchr(linebuf, '\r')) != NULL)
                        *strptr = '\0';


		    /* Get device find output parameters from linebuf. */
		    strv = strexp(linebuf, &strc);

		    /* POSIX standard expects 6 parameters on each line. */
		    if(strc >= 6)
		    {
			strncpy(dev, strv[0], PATH_MAX + NAME_MAX);
			dev[PATH_MAX + NAME_MAX - 1] = '\0';

			capacity_total = atol(strv[1]);
			capacity_used = atol(strv[2]);
			capacity_available = atol(strv[3]);

			/* Calculate capacity percent. */
			capacity_percent = (double)capacity_used /
			    (double)capacity_total * 100;

                        strncpy(path, strv[5], PATH_MAX + NAME_MAX);
                        path[PATH_MAX + NAME_MAX - 1] = '\0';

			/*   See which devrec above information belongs to.
                         *   Match by mounted to path.
			 */
			for(i = 0; i < total_devrecs; i++)
			{
                            if(!DevRecIsAllocated(i)) continue;
			    if(devrec[i]->path == NULL) continue;
/*
printf("`%s' `%s'\n", devrec[i]->path, path);
*/
			    if(strcmp(devrec[i]->path, path)) continue;

			    /* Set new capacity values to device record. */
			    devrec[i]->capacity_total = capacity_total;
                            devrec[i]->capacity_used = capacity_used;
                            devrec[i]->capacity_available = capacity_available;
                            devrec[i]->capacity_percent = capacity_percent;
/*
printf("%s: %i %i %i %.0lf\n",
dev,
devrec[i]->capacity_total,
devrec[i]->capacity_used,
devrec[i]->capacity_available,
devrec[i]->capacity_percent
);
*/
			}
		    }

		    /* Free parameters. */
		    for(i = 0; i < strc; i++)
		        free(strv[i]);
		    free(strv); strv = NULL;
		    strc = 0;

		    /* Get next line. */
		    free(linebuf); linebuf = NULL;
		    linebuf = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
		}
		/* Free line buffer. */
		free(linebuf); linebuf = NULL;

		/* Close file. */
		fclose(fp); fp = NULL;

	    }	/* Open and read output file. */
	}


	return(0);
}



/*
 *	Allocates a new device record.
 */
int DevRecAllocate()
{
	int i;
	int new_rec_num;


	new_rec_num = -1;

	if((int)total_devrecs < 0)
            total_devrecs = 0;

	/* Check for NULL pointer in existing scope of array. */
	for(i = 0; i < (int)total_devrecs; i++)
	{
	    if(devrec[i] == NULL)
		break;
	}
	/* Did we find an available array pointer. */
	if(i < (int)total_devrecs)
	{
	    new_rec_num = i;
	}
	/* Else we need to allocate more pointers. */
	else
	{
	    new_rec_num = total_devrecs;
	    total_devrecs++;	/* Increment global variable total. */

	    devrec = (devrec_struct **)realloc(devrec,
		total_devrecs * sizeof(devrec_struct *)
	    );
	    if(devrec == NULL)
	    {
		total_devrecs = 0;
		return(-1);
	    }
	}


	/* Allocate new structure. */
	devrec[new_rec_num] = (devrec_struct *)calloc(1, sizeof(devrec_struct));
        if(devrec[new_rec_num] == NULL)
            return(-1);


	return(new_rec_num);
}



/*
 *	Deletes device record rec_num.
 */
void DevRecDelete(int rec_num)
{
	if(!DevRecIsAllocated(rec_num))
	    return;


	free(devrec[rec_num]->label);
	devrec[rec_num]->label = NULL;
	free(devrec[rec_num]->devname);
	devrec[rec_num]->devname = NULL;
        free(devrec[rec_num]->path);
        devrec[rec_num]->path = NULL;
        free(devrec[rec_num]->filesystem);
        devrec[rec_num]->filesystem = NULL;

        free(devrec[rec_num]->cmd_mount);
        devrec[rec_num]->cmd_mount = NULL;
        free(devrec[rec_num]->cmd_unmount);
        devrec[rec_num]->cmd_unmount = NULL;
        free(devrec[rec_num]->cmd_status);
        devrec[rec_num]->cmd_status = NULL;


	/* Free structure itself. */
	free(devrec[rec_num]);
	devrec[rec_num] = NULL;


	return;
}



void DevRecDeleteAllRecords()
{
	int i;

	for(i = 0; i < (int)total_devrecs; i++)
	{
	    DevRecDelete(i);
	}


	free(devrec);
	devrec = NULL;

	total_devrecs = 0;


	return;
}



/* **********************************************************************
 *
 *
 */
void DevRecDoComfUnmountAll()
{
	int i;
	int button_code;


	/* Do we want comfermation? */
	if(!option.comferm_unmount_onexit)
	    return;

	/* Move comfermation dialog to pointer's center. */
	WidgetCenterWindow(comf_dialog.toplevel, WidgetCenterWindowToPointer);

	/* Turn yes to all global off. */
	yes_to_all = False;

	for(i = 0; i < total_devrecs; i++)
	{
	    if(!DevRecIsAllocated(i)) continue;
	    if(!DevRecIsReallyMounted(devrec[i]->devname)) continue;
	    if(devrec[i]->flags & DEV_TF_NOUNMOUNT) continue;

	    /* Check comfermation of device unmount. */
	    if(!yes_to_all)
	    {
	        button_code = ComfDialogMapQueryUnmount(devrec[i]->label);
	        if(button_code == NO_PBTN_NUM)
		    continue;
		else if(button_code == ALL_PBTN_NUM)
		    yes_to_all = True;
	    }

	    /* Unmount the device. */
	    DevRecDoUnmount(i);
	}

        /* Turn yes to all global off. */
        yes_to_all = False;



	return;
}
