/*
                        Endeavour Browser: Callbacks

	Functions:

        int EDWPerformFileOp(
                int op_code
                char *tar_path,
                char *src_path
        )
        int EDWProcessMovement(
                void *tar_ptr, int tar_ptr_type,
                void *src_ptr, int src_ptr_type,  
                int tar_drag_type, int src_drag_type,
                int tar_item_num,
                int op_code,
                bool_t comfermed
        )
        int EDWExecuteSelectedFiles(endeavour_win_struct *edw, bool_t execute)
	int EDWMenuCB(void *client_data, int id_code)

	---



 */

#include "endeavour.h"  
#include "edwdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))



/*
 *	Performs a file operation and updates Endeavour Browsers.
 */
int EDWPerformFileOp(
        int op_code,
        char *tar_path,
        char *src_path
)
{
        int status;
        int i, n;
        struct stat stat_buf;
	char stringa[PATH_MAX + NAME_MAX + 128];


        /*      If both paths are the same, perform no operation.
         *      With the exception of the following operations:
         *
         *      Recycle, Recover.
         */
        if((tar_path != NULL) && (src_path != NULL))
        {
            if(!strcmp(tar_path, src_path) &&
               (op_code != OP_CODE_RECYCLE) &&
               (op_code != OP_CODE_RECOVER) &&
               (op_code != OP_CODE_CHOWN_OWNER) &&
               (op_code != OP_CODE_CHOWN_GROUP)
            )
                return(0);
        }


        /* ************************************************************* */

        status = 0;
        switch(op_code)
        {
          /* ************************************************************* */
          case OP_CODE_COPY:
            /* Error checks. */
            if((tar_path == NULL) ||
               (src_path == NULL)
            )
                return(-1);

            SimplifyPath(tar_path);
            SimplifyPath(src_path);


            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw, 
"Cannot copy object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"   
                );
                return(-3);
            }
        
            /* Both arguments must be absolute path. */
            if(!ISPATHABSOLUTE(tar_path) ||
               !ISPATHABSOLUTE(src_path)
            )
	    {
                fprintf(stderr,
		    "EDWPerformFileOp(): Input path not absolute.\n"
		);
                return(-2);
            }

            /* Source must exist. */
            if(lstat(src_path, &stat_buf))
                return(-2);
         
            /* Check if target exists and if we can overwrite. */
            if(DESTEXISTS(tar_path, src_path))
            {
                if(option.comferm_overwrites &&
                   !yes_to_all   
                )
                {
                    i = ComfDialogMapQueryOverWrite(
                        tar_path,  
                        src_path,  
                        op_code
                    );
                    if(i == NO_PBTN_NUM)
                        return(-4); 
                    else if(i == ALL_PBTN_NUM)
                    {
                        /* Switch yes to all on. */
                        yes_to_all = True;
                    }

                } 
                
                /* Remove file first. */
                if(!ISPATHDIR(tar_path))
                    unlink(tar_path);
            }
               
            /* Copy file or directory by what the source is. */
            if(ISPATHDIR(src_path))
	    {
                FileOpCopyDir(tar_path, src_path);

                /* Update listings on all edws. */ 
                for(i = 0; i < total_endeavour_wins; i++)
                {
                    if(endeavour_win[i] == NULL) continue;

                    /* Update target listing. */
/*
                    n = EDWMatchExistingDirFromPath(endeavour_win[i], tar_path);
                    EDWColapseDir(endeavour_win[i], n);
                    EDWExpandDir(endeavour_win[i], n);
*/
                }
            }
            else
            {
                FileOpCopyFile(tar_path, src_path);

                /* Update listings on all edws. */   
                for(i = 0; i < total_endeavour_wins; i++)
                {
                    if(endeavour_win[i] == NULL) continue;

                    /* Update target listing. */
/*
                    n = EDWMatchExistingDirFromPath(endeavour_win[i], tar_path);
                    EDWColapseDir(endeavour_win[i], n);
                    EDWExpandDir(endeavour_win[i], n);
*/
                }
	    }
            break;

          /* ************************************************************* */
          case OP_CODE_MOVE:
            /* Error checks. */    
            if((tar_path == NULL) ||
               (src_path == NULL)
            )
                return(-1);

            SimplifyPath(tar_path);
            SimplifyPath(src_path);


            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot move object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }

            /* Both arguments must be absolute path. */
            if(!ISPATHABSOLUTE(tar_path) ||
               !ISPATHABSOLUTE(src_path)
            )
	    {
                fprintf(stderr,
                    "EDWPerformFileOp(): Input path not absolute.\n"
                );
                return(-2);
	    }

            /* Source must exist locally. */
            if(lstat(src_path, &stat_buf))
                return(-2);

            /* Check if target exists and if we can overwrite. */
            if(DESTEXISTS(tar_path, src_path))
            {
                /* Prompt for overwrite. */
                if(option.comferm_overwrites &&
                   !yes_to_all
                )
                {
                    i = ComfDialogMapQueryOverWrite(
                        tar_path,  
                        src_path,
                        op_code
                    );
                    if(i == NO_PBTN_NUM)
                        return(-4);
                    else if(i == ALL_PBTN_NUM)
                    {
                        /* Switch yes to all on. */
                        yes_to_all = True;
                    }
                } 
                
                /* Remove file first. */
                if(!ISPATHDIR(tar_path))
                    unlink(tar_path);
            }

            /* Move file or directory. */
            if(ISPATHDIR(src_path))
	    {
                FileOpMoveDir(tar_path, src_path);

                /* Update listings on all edws. */
                for(i = 0; i < total_endeavour_wins; i++)
                {
                    if(endeavour_win[i] == NULL) continue;

                    /* Delete any source entries that exist in listed memory. */
                    n = EDWMatchExistingDirFromPath(endeavour_win[i], src_path);
                    EDWColapseDir(endeavour_win[i], n);
                    EDWDeleteDirItemPath(endeavour_win[i], src_path);

                    /* Update target listing. */
/*
                    n = EDWMatchExistingDirFromPath(endeavour_win[i], tar_path);
                    EDWColapseDir(endeavour_win[i], n);
                    EDWExpandDir(endeavour_win[i], n);
*/
                }
	    }
            else
	    {
                FileOpMoveFile(tar_path, src_path);

	        /* Update listings on all edws. */
                for(i = 0; i < total_endeavour_wins; i++)
                {
		    if(endeavour_win[i] == NULL) continue;

		    /* Update target listing. */
/*
                    n = EDWMatchExistingDirFromPath(endeavour_win[i], tar_path);
                    EDWColapseDir(endeavour_win[i], n);
		    EDWExpandDir(endeavour_win[i], n);
*/
		}
            }
            break;

          /* ************************************************************* */
          case OP_CODE_RENAME:
            /* Error checks. */
            if((tar_path == NULL) ||
               (src_path == NULL) 
            )
                return(-1);
                      
            SimplifyPath(tar_path);
            SimplifyPath(src_path);
                    
                     
            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot rename object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }
            
            /* Both arguments must be absolute path. */
            if(!ISPATHABSOLUTE(tar_path) ||
               !ISPATHABSOLUTE(src_path)
            )
                return(-2);
                
            /* Source must exist. */
            if(lstat(src_path, &stat_buf))
                return(-2);

            /* Check if target exists. */
            if(lstat(tar_path, &stat_buf) == 0)
	    {
		sprintf(stringa,
"Cannot rename object:\n\n\
    `%s' already exists.\n\n",
		    tar_path
		);
		printdw(&err_dw, stringa);

                return(-2);
	    }

            /* Rename file or directory. */
            if(ISPATHDIR(src_path))
                FileOpRenameDir(tar_path, src_path);
            else
                FileOpRenameFile(tar_path, src_path);
	    /* src_path no longer exists on disk after this point. */

            /* Update listings on all edws. */
            for(i = 0; i < total_endeavour_wins; i++)
            {
                if(endeavour_win[i] == NULL) continue;

		if(ISPATHDIR(tar_path))
		{
                    /* Update dir item name on the list. */
                    EDWUpdateDirItemName(
                        endeavour_win[i],
                        src_path,	/* Old name. */
                        tar_path	/* New name. */
                    );
                }
                else
                {
                    /* Update file listing. */
                    EDWRefreshFileList(
                        endeavour_win[i],
                        endeavour_win[i]->loc_prompt.buf
                    );
                }
            }
            break;
                
          /* ************************************************************* */
          case OP_CODE_RECYCLE:
            /* Error checks. */
            if(src_path == NULL)        /* Target path is not used. */
                return(-1);

            SimplifyPath(src_path);

            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot recycle object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }

            /* Source must be an absolute path. */
            if(!ISPATHABSOLUTE(src_path))
                return(-2);

            /* Source must exist, use lstat to check if it is a link. */
            if(lstat(src_path, &stat_buf))
                return(-1);


            /* Prompt for recycle comfermation. */
            if(option.comferm_recycle &&
               !yes_to_all
            )   
            {
                i = ComfDialogMapQueryOverWrite(
                    NULL,
                    src_path,
                    op_code
                );
                if(i == NO_PBTN_NUM)
                    return(-4);
                else if(i == ALL_PBTN_NUM)
                {
                    /* Switch yes to all on. */
                    yes_to_all = True;
                }
            }

            /* Recycle the object. */
            RecycleDoRecycleObject(src_path, fname.recycle_bin);

            /* Redraw recycle bin window if mapped. */
            if((int)total_recbin_wins > 0)
            {
                if(recbin_win[0] != NULL)
                {
                    RECBINREFRESHLISTING(recbin_win[0]);
                    if(recbin_win[0]->map_state == 1)
                    {
                        RecBinDraw(recbin_win[0], DRAW_AMOUNT_COMPLETE);
                    }
                }
            }   
            break;
               
          /* ************************************************************* */
          case OP_CODE_RECOVER:
            if((tar_path == NULL) ||
               (src_path == NULL)
            )
                return(-1);
                
            /* Simplify only target path. */
            SimplifyPath(tar_path);
                 
                    
            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot recover object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }
                 
            /*   src_path is the item in the recbin we want to recover.
             *   tar_path is where we want to recover to.
             */
            RecycleDoRestoreObject(
                src_path,
                fname.recycle_bin,
                tar_path
            );
            
            break;
          
          /* ************************************************************* */
          case OP_CODE_CHOWN_OWNER:
            if((tar_path == NULL) ||
               (src_path == NULL)
            )
                return(-1);

            /* Source must exist, use lstat to check if it is a link. */
            if(lstat(src_path, &stat_buf))
                return(-2);

            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot chown object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-2);
            }    
            
            FileOpChownOwner(
                src_path,       /* Source object. */
                tar_path        /* Argument. */
            );


            break;
            
          /* ************************************************************* */
          case OP_CODE_CHOWN_GROUP:
            if((tar_path == NULL) ||
               (src_path == NULL)   
            )
                return(-1);
                
            /* Source must exist, use lstat to check if it is a link. */
            if(lstat(src_path, &stat_buf))
                return(-2);
                
            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot chown object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }
                 
            FileOpChownGroup(
                src_path,       /* Source object. */
                tar_path        /* Argument. */
            );
              
             
            break;
              
          /* ************************************************************* */
          case OP_CODE_TEST:
            fprintf(stderr,
 "EDWPerformFileOp(): TEST: Operation starting...\n"
            );

            if((tar_path == NULL) ||
               (src_path == NULL)
            )
	    {
                fprintf(stderr,
 "EDWPerformFileOp(): TEST: One or more object paths is a NULL pointer.\n"
                );
                return(-1);
            }

            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                fprintf(stderr,
 "EDWPerformFileOp(): TEST: Master write protect is on, operation aborted.\n"
                );
                return(-3);
            }
             
            /* Both arguments must be absolute path. */
            if(!ISPATHABSOLUTE(tar_path) ||
               !ISPATHABSOLUTE(src_path)
            )
	    {
                fprintf(stderr,
 "EDWPerformFileOp(): TEST: One or more input path(s) is not absolute.\n"
                );
                return(-2);
            }

            /* Source must exist. */
            if(lstat(src_path, &stat_buf))
	    {
		fprintf(stderr,
 "EDWPerformFileOp(): TEST: %s: Source object does not exist (local).\n",
		    src_path
                );
                return(-2);
            }

            /* Check if target exists and if we can overwrite. */
            if(DESTEXISTS(tar_path, src_path)) 
            {
                fprintf(stderr,
 "EDWPerformFileOp(): TEST: Warning: Destination exists.\n"
		);

                /* Prompt for overwrite. */
                if(option.comferm_overwrites)
                {
                    i = ComfDialogMapQueryOverWrite(
                        tar_path,
                        src_path,
                        op_code   
                    );
                    if(i == NO_PBTN_NUM)
                        return(-4);
                    else if(i == ALL_PBTN_NUM)
                    {
                        /* Do not switch yes to all on for test. */

                    }
                }
             
                /* Remove file first. */
                if(!ISPATHDIR(tar_path))
		{
                    fprintf(stderr,
 "EDWPerformFileOp(): TEST: %s: Would have been removed.\n",
			tar_path
		    );
		}
            }

            fprintf(stderr,
 "EDWPerformFileOp(): TEST: Operation executed completely.\n"
            );

            break;
                  
          /* ************************************************************* */
          default:
            printdw(&err_dw,
"Operation code not supported.\n\n\
(This is not a critical error.)\n"
            );
            break;
        }
             
                
              
                 
        return(status);
}



/*
 *      Handles drag movements.   If comfermation is needed, and
 *      comfermed is False, then the request will be sent to the
 *      Comfermation Dialog to handle the movement.
 */
int EDWProcessMovement(
        void *tar_ptr,          /* Always an edw. */
        int tar_ptr_type,       /* Must be POINTER_TYPE_EDW. */
        void *src_ptr,
        int src_ptr_type,
        
        int tar_drag_type,
        int src_drag_type,
        int tar_item_num,
        int op_code,
        bool_t comfermed
)
{
        int i, n; 
/*
        int tar_file_num, tar_dir_num;
        int src_file_num, src_dir_num;
 */

        char tmp_path[PATH_MAX + NAME_MAX];
        char tar_path[PATH_MAX + NAME_MAX];
        char src_path[PATH_MAX + NAME_MAX];
        char tmp_name[PATH_MAX + NAME_MAX];
    
        char need_comferm;


        endeavour_win_struct *src_edw, *tar_edw;
        recbin_win_struct *src_rb;
#ifdef USE_IMLIB
        img_viewer_struct *src_iv;
        img_browser_struct *src_ib;
#endif /* USE_IMLIB */


        /* Error checks. */
        if((tar_ptr == NULL) ||         /* Target must exist. */
           (tar_ptr_type == POINTER_TYPE_UNDEFINED) 
        )
            return(-1);


        /* Target must be endeavour window. */
        if(tar_ptr_type != POINTER_TYPE_EDW)
        {
            fprintf(stderr,
        "EDWProcessMovement(): tar_ptr_type must be POINTER_TYPE_EDW.\n"
            );
            return(-1);
        }


        /* ************************************************************ */
        /* Get target and source pointers. */
        tar_edw = (endeavour_win_struct *)tar_ptr;      /* Always not NULL. */
        src_edw = NULL;
        src_rb = NULL;

#ifdef USE_IMLIB
        src_iv = NULL;
        src_ib = NULL;
#endif /* USE_IMLIB */

        /* Check source pointer type. */
        switch(src_ptr_type)
        {
          case POINTER_TYPE_EDW:
            src_edw = (endeavour_win_struct *)src_ptr;
            break;

          case POINTER_TYPE_RB:
            src_rb = (recbin_win_struct *)src_ptr;  
            break;

#ifdef USE_IMLIB
          case POINTER_TYPE_IV:
            src_iv = (img_viewer_struct *)src_ptr;
            break;

          case POINTER_TYPE_IB:
            src_ib = (img_browser_struct *)src_ptr;
            break;
#endif /* USE_IMLIB */ 

          default:
            fprintf(stderr,
        "EDWProcessMovement(): Unknown source pointer type %i.\n",
                src_ptr_type
            );
            return(-1);
            break;
        }

        /* ************************************************************ */
        /* Endeavour Browser to Endeavour Browser? */
        if((src_edw != NULL) &&
           (tar_edw != NULL)
        )
        {
            /* Both edw location prompts must contain data. */
            if(tar_edw->loc_prompt.buf == NULL)
                return(-1);
            if(src_edw->loc_prompt.buf == NULL)
                return(-1);
          
            /* Duplicity check: Dir list to dir list on same edw. */
            if((src_edw == tar_edw) &&
               (tar_drag_type == DRAG_TYPE_DIR) &&
               (src_drag_type == DRAG_TYPE_DIR)
            )
            {
                if((int)src_edw->total_sel_dirs > 0)
                {
                    /* Get first selected dir on source edw. */
                    n = src_edw->sel_dir[0];
                    /* Is tar_item_num the same num as the one on the src? */
                    if(n == tar_item_num)
		    {
                        EDWRefreshAllLists(src_edw);
                        return(0);
		    }
                }
            }
            /* Duplicity check: File to file, check if going to same dir. */
            if((src_edw == tar_edw) &&
               (tar_drag_type == DRAG_TYPE_FILE) &&
               (src_drag_type == DRAG_TYPE_FILE)
            )
            {
                return(0);
            }
            if((tar_drag_type == DRAG_TYPE_FILE) &&  
               (src_drag_type == DRAG_TYPE_FILE)
            )
            {
                /* Are the current locations the same on both edws? */
                if(!strcmp(src_edw->loc_prompt.buf, tar_edw->loc_prompt.buf))
                {
                    return(0);
                }
            }
          
            /* Comfermation needed? */
            if(!comfermed)
            {
                if((op_code == OP_CODE_COPY) &&
                   option.comferm_pointer_motion_copy
                )
                    need_comferm = 1;
                else if((op_code == OP_CODE_MOVE) &&
                        option.comferm_pointer_motion_move
                )
                    need_comferm = 1;
                else
                    need_comferm = 0;
                  
                if(need_comferm)
                {
                    /* Move comfermation dialog to pointer's center. */
                    WidgetCenterWindow(comf_dialog.toplevel,
                        WidgetCenterWindowToPointer);
             
                    ComfDialogMapQueryMovement(
                        tar_edw, tar_ptr_type,
                        src_edw, src_ptr_type,
                        tar_drag_type,
                        src_drag_type,
                        tar_item_num,
                        op_code
                    );
                
                    /*   Return, do not continue.
                     *   Comfermation will handle query and call this function
                     *   again as needed.
                     */
                    return(0);
                }
            }
             
            /* Target drag type to directory list? */
            if(tar_drag_type == DRAG_TYPE_DIR)
            {
                /* Check if tar_item_num is valid. */
                if((tar_item_num < 0) ||
                   (tar_item_num >= tar_edw->total_dirs)  
                )
                    return(-1);
                if(tar_edw->dir[tar_item_num] == NULL)
                    return(-1);
                if(tar_edw->dir[tar_item_num]->fullname == NULL)
                    return(-1); 
                 
                /* Get tar_path. */
                strncpy(
                    tar_path,
                    tar_edw->dir[tar_item_num]->fullname,
                    PATH_MAX + NAME_MAX
                );
            }
            /* Target drag type to file list? */
            else if(tar_drag_type == DRAG_TYPE_FILE)
            {
                /* No need to validate tar_item_num. */
                      
                /* Get tar_path from prompt buf. */
                strncpy(
                    tar_path,
                    tar_edw->loc_prompt.buf,
                    PATH_MAX + NAME_MAX
                );
            }
            else
            {
                /* Unsupported target drag type. */  
                printdw(&err_dw,
"Internal non-critical error:\n\n\
    Unsupported target drag type.\n\n"
                );
                return(-2);
            }
            tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
            SimplifyPath(tar_path);
         
            /* Move comfermation dialog to pointer's center. */ 
            WidgetCenterWindow(comf_dialog.toplevel,
                WidgetCenterWindowToPointer);
                
                
            /* Source drag type file? */
            if(src_drag_type == DRAG_TYPE_FILE)
            {
                for(i = 0; i < (int)src_edw->total_sel_files; i++)
                {
                    n = src_edw->sel_file[i];   
                    if((n < 0) || (n >= (int)src_edw->total_files))
                        continue;
                    if(src_edw->file[n] == NULL) continue;
                    if(src_edw->file[n]->name == NULL) continue;
                
                    /* Get source path. */
                    strncpy(src_path,
                        PrefixPaths(
                            src_edw->loc_prompt.buf,
                            src_edw->file[n]->name
                        ),
                        PATH_MAX + NAME_MAX
                    );
                    src_path[PATH_MAX + NAME_MAX - 1] = '\0';
                    SimplifyPath(src_path);

                    /* Skip if paths are the same. */
                    if(!strcmp(tar_path, src_path))
                        continue;
             
                    /* Perform file operation. */
                    EDWPerformFileOp(op_code, tar_path, src_path);
                }
            }
            /* Source drag type directory? */
            else if(src_drag_type == DRAG_TYPE_DIR)
            {   
                for(i = 0; i < (int)src_edw->total_sel_dirs; i++)
                {
                    n = src_edw->sel_dir[i];   
                    if((n < 0) || (n >= (int)src_edw->total_dirs))
                        continue;
                    if(src_edw->dir[n] == NULL) continue;
                    if(src_edw->dir[n]->fullname == NULL) continue;

                    /* Get source path. */
                    strncpy(
                        src_path,
                        src_edw->dir[n]->fullname,
                        PATH_MAX + NAME_MAX
                    );
                    src_path[PATH_MAX + NAME_MAX - 1] = '\0';
                    SimplifyPath(src_path);

                    /* Perform file operation. */
                    EDWPerformFileOp(op_code, tar_path, src_path);
                }
            }
	    /* tar_item_num might be invalid now. */

	    EDW_UNSELECTALLDIRS(tar_edw);
	    EDW_UNSELECTALLFILES(tar_edw);
	    EDW_UNSELECTALLDIRS(src_edw);
            EDW_UNSELECTALLFILES(src_edw);

	    /* Reget directory listing if target is dir. */
            if(src_drag_type == DRAG_TYPE_DIR)
            {
		/* Remember, tar_item_num might be invalid now. */
		n = EDWMatchExistingDirFromPath(tar_edw, tar_path);
		EDWColapseDir(tar_edw, n);
                EDWExpandDir(tar_edw, n);
  	    }
	    if(tar_edw != src_edw)
	    {
                /* Remember, tar_item_num might be invalid now. */
                n = EDWMatchExistingDirFromPath(src_edw, tar_path);
                EDWColapseDir(src_edw, n);
                EDWExpandDir(src_edw, n); 
	    }

            /* Refresh listings. */
            EDWRefreshAllLists(tar_edw);
            if(tar_edw != src_edw)
                EDWRefreshAllLists(src_edw);

            EDWDraw(tar_edw, DRAW_AMOUNT_COMPLETE);
            if(tar_edw != src_edw)
                EDWDraw(src_edw, DRAW_AMOUNT_COMPLETE);
        }
        /* ********************************************************** */
        /* Recycle Bin to Endeavour Browser? */
        else if((src_rb != NULL) &&
                (tar_edw != NULL)
        )
        {
            /* Check if comfermation is needed. */
            if(!comfermed)
            {
                if((op_code == OP_CODE_COPY) &&
                   option.comferm_pointer_motion_copy
                )
                    need_comferm = 1;
                else if((op_code == OP_CODE_MOVE) &&
                        option.comferm_pointer_motion_move
                )           
                    need_comferm = 1;
                else
                    need_comferm = 0;
                    
                if(need_comferm)
                {
                    /* Move comfermation dialog to pointer's center. */
                    WidgetCenterWindow(comf_dialog.toplevel,
                        WidgetCenterWindowToPointer);
            
                    ComfDialogMapQueryMovement(
                        tar_edw, tar_ptr_type,
                        src_rb, src_ptr_type,
                        tar_drag_type,
                        src_drag_type,
                        tar_item_num,
                        OP_CODE_RECOVER         /* This needs to be recover. */
                    );
                    /*   Return, do not continue.
                     *   Comfermation will handle query and call this function
                     *   again as needed.
                     */
                    return(0);
                }
            }
            
            /* Location prompts must contain data. */
            if(tar_edw->loc_prompt.buf == NULL)
                return(-1);
                 
            /* Target type directory. */
            if(tar_drag_type == DRAG_TYPE_DIR)
            {
                /* Check if tar_item_num is valid. */
                if((tar_item_num < 0) ||
                   (tar_item_num >= (int)tar_edw->total_dirs)
                )
                    return(-1);
                if(tar_edw->dir[tar_item_num] == NULL)
                    return(-1);
                if(tar_edw->dir[tar_item_num]->fullname == NULL)
                    return(-1);
                        
                /* Get tar_path. */
                strncpy(
                    tar_path,
                    tar_edw->dir[tar_item_num]->fullname,
                    PATH_MAX + NAME_MAX
                );
                SimplifyPath(tar_path);
            }
            /* Target type file. */
            else if(tar_drag_type == DRAG_TYPE_FILE)
            {
                /* No need to check if tar_item_num is valid. */
                       
                /* Get tar_path. */
                strncpy(
                    tar_path,
                    tar_edw->loc_prompt.buf,
                    PATH_MAX + NAME_MAX
                );
                SimplifyPath(tar_path);
            }    
            else
            {
                /* Unsupported target drag type. */
                printdw(&err_dw,
"Internal non-critical error:\n\n\
    Unsupported target drag type.\n\n"
                );
                return(-2);
            }
                    
                
            /* Move comfermation dialog to pointer's center. */
            WidgetCenterWindow(comf_dialog.toplevel,
                WidgetCenterWindowToPointer);
                
                    
            /*   Now we have the tar_path, this is where we want to
             *   recover recycled objects to.
             */   
            for(i = 0; i < (int)src_rb->total_sel_objects; i++)
            {
                n = src_rb->sel_object[i];
                if((n < 0) || (n >= (int)src_rb->total_objects))
                    continue;
                if(src_rb->object[n] == NULL) continue;
                if(src_rb->object[n]->name == NULL) continue;
                
                /* Get source path (in the recbin). */
                strncpy(
                    src_path,
                    src_rb->object[n]->name,
                    PATH_MAX + NAME_MAX
                );
                src_path[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Get tmp name, remove its full path. */
                strncpy(tmp_name, src_rb->object[n]->name,
                    PATH_MAX + NAME_MAX);
                tmp_name[PATH_MAX + NAME_MAX - 1] = '\0';
                StripAbsolutePath(tmp_name);
                  
                /* Prefix the target path to the tmp_name. */
                strncpy(tmp_path, PrefixPaths(tar_path, tmp_name),
                    PATH_MAX + NAME_MAX);
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
                SimplifyPath(tmp_path);
            
                /* Recover object from recycle bin. */
                EDWPerformFileOp(
                    OP_CODE_RECOVER,    /* Recbin to edw is always recover. */
                    tmp_path,           /* Full path name to recover to. */
                    src_path            /* Object in recbin. */
                );
            }
            /* tar_item_num might be invalid now. */


            EDW_UNSELECTALLDIRS(tar_edw);
            EDW_UNSELECTALLFILES(tar_edw);

            /* Reget directory listing if target is dir. */
            if((tar_drag_type == DRAG_TYPE_DIR)
/* Check if source was a dir? */
            )
            {
                n = EDWMatchExistingDirFromPath(tar_edw, tar_path);
                EDWColapseDir(tar_edw, n);
                EDWExpandDir(tar_edw, n);
            }

            /* First refresh recycle bin listing. */
            RECBINREFRESHLISTING(src_rb);
            RecBinDraw(src_rb, DRAW_AMOUNT_COMPLETE);  

            /* Next refresh edw listing after. */
            EDWRefreshAllLists(tar_edw);
            EDWDraw(tar_edw, DRAW_AMOUNT_COMPLETE);
        }
#ifdef USE_IMLIB
        /* *************************************************************** */
        /* Image Browser to Endeavour Browser? */
        else if((src_ib != NULL) &&
                (tar_edw != NULL)
        )
        {
            /* Edw location prompts must contain data. */
            if(tar_edw->loc_prompt.buf == NULL)
                return(-1);
            if(src_ib->current_dir == NULL)
                return(-1);
                
            /* Check if comfermation is needed. */
            if(!comfermed)
            {
                if((op_code == OP_CODE_COPY) &&
                   option.comferm_pointer_motion_copy 
                )
                    need_comferm = 1;
                else if((op_code == OP_CODE_MOVE) &&
                        option.comferm_pointer_motion_move
                ) 
                    need_comferm = 1;
                else
                    need_comferm = 0;
            
                if(need_comferm)
                {
                    /* Move comfermation dialog to pointer's center. */
                    WidgetCenterWindow(comf_dialog.toplevel,
                        WidgetCenterWindowToPointer);
            
                    ComfDialogMapQueryMovement(
                        tar_edw, tar_ptr_type,
                        src_ib, src_ptr_type,
                        tar_drag_type,
                        src_drag_type,
                        tar_item_num,
                        op_code
                    );
          
                    /*   Return, do not continue.
                     *   Comfermation will handle query and call this function
                     *   again as needed.  
                     */
                    return(0);
                }
            }
             
            /* Target type directory? */
            if(tar_drag_type == DRAG_TYPE_DIR)
            {
                /* Check if tar_item_num is valid. */
                if((tar_item_num < 0) ||
                   (tar_item_num >= (int)tar_edw->total_dirs)
                ) 
                    return(-1);
                if(tar_edw->dir[tar_item_num] == NULL)
                    return(-1);
                if(tar_edw->dir[tar_item_num]->fullname == NULL)
                    return(-1); 
                 
                /* Get tar_path. */
                strncpy(
                    tar_path,
                    tar_edw->dir[tar_item_num]->fullname,
                    PATH_MAX + NAME_MAX
                );
                tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
                SimplifyPath(tar_path);
            }
            /* Target type file? */  
            else if(tar_drag_type == DRAG_TYPE_FILE)
            {
                /* Get tar_path. */
                strncpy(tar_path, tar_edw->loc_prompt.buf, PATH_MAX + NAME_MAX);
                tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
                SimplifyPath(tar_path);    
            }
            else
            {
                /* Unsupported target drag type. */
                printdw(&err_dw,
"Internal non-critical error:\n\n\
    Unsupported target drag type.\n\n"
                );
                return(-2);

            }
                   
            /* Move comfermation dialog to pointer's center. */
            WidgetCenterWindow(comf_dialog.toplevel,
                WidgetCenterWindowToPointer);
                
            /* Process items on source Image browser. */
            for(i = 0; i < (int)src_ib->total_sel_thumbs; i++)
            {    
                n = src_ib->sel_thumb[i];
                if((n < 0) || (n >= (int)src_ib->total_thumbs))
                    continue;
                if(src_ib->thumb[n] == NULL) continue;   
                if(src_ib->thumb[n]->filename == NULL) continue;
                  
                /* Get source path. */
                strncpy(src_path,
                    PrefixPaths(
                        src_ib->current_dir,
                        src_ib->thumb[n]->filename  
                    ),
                    PATH_MAX + NAME_MAX
                );
                src_path[PATH_MAX + NAME_MAX - 1] = '\0';
                SimplifyPath(src_path);    

                EDWPerformFileOp(op_code, tar_path, src_path);
            }
            /* tar_item_num might be invalid now. */

            EDW_UNSELECTALLDIRS(tar_edw);
            EDW_UNSELECTALLFILES(tar_edw);

            /* Reget directory listing if target is dir. */
            if((tar_drag_type == DRAG_TYPE_DIR)
/* Check if source was a dir? */
            )
            {
                n = EDWMatchExistingDirFromPath(tar_edw, tar_path);
                EDWColapseDir(tar_edw, n);
                EDWExpandDir(tar_edw, n);
            }
                  
            /* First refresh image browser thumb listing. */
            strncpy(tmp_path, src_ib->current_dir, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
            IMBRLoad(src_ib, tmp_path);
            IMBRDraw(src_ib, DRAW_AMOUNT_COMPLETE); 
             
            /* Next refresh edw listing after. */
            EDWRefreshAllLists(tar_edw);
            EDWDraw(tar_edw, DRAW_AMOUNT_COMPLETE);
        }
#endif /* USE_IMLIB */
                
                    
        /* Turn yes to all off. */
        yes_to_all = False;
                  
                
        return(0);
}



/*
 *      Executes selected files by extension.
 *       
 *      If execute is set to False, then nothing will be executed and the
 *      cmd_prompt buffer is placed with the command line that would have
 &      been executed.
 */
int EDWExecuteSelectedFiles(endeavour_win_struct *edw, bool_t execute)
{
        int executions;
        int sel_num, file_num;

        char cmd[PATH_MAX + NAME_MAX + 256];
        char tmp_path[PATH_MAX + NAME_MAX];
        char *strptr;


        /* Error checks. */
        if(edw == NULL)
            return(-1);

        if(edw->total_sel_files < 1)
            return(-1);

        if(edw->loc_prompt.buf == NULL)
            return(-1);

        /* Change current working directory. */
        if(chdir(edw->loc_prompt.buf) == 0)
        {
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
        }


        /* ******************************************************** */   
        executions = 0;

        for(sel_num = 0; sel_num < edw->total_sel_files; sel_num++)
        {
            /* Don't execute too many. */
            if(executions >= option.file_exe_max_burst)
                break;

            /* Get file number from selection. */
            file_num = edw->sel_file[sel_num];

            /* Check if file number is valid. */
            if((file_num < 0) || (file_num >= edw->total_files))
                continue;
            if(edw->file[file_num] == NULL)
		continue;
            if(edw->file[file_num]->name == NULL)
		continue;


            /* Get tmp_path as absolute path and name of the object. */
            strptr = PrefixPaths(
                edw->loc_prompt.buf,
                edw->file[file_num]->name
            );
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
        
            /*   NOTE: Check for extension, object maybe on a filesystem
             *   of which everything is set executeable.  So only if there
             *   is no extension, then check if it is executeable.
             */
            if(FILEHASEXTENSION(edw->file[file_num]->name))
            {
                /* Has extension, search through file associations for the
                 * appropriate command.
                 */
                strptr = FAssGetCmdByExt(
                    FAssExtParse(
                        edw->file[file_num]->name
                    )
                );
                /* Does it have an extension? */
                if(strptr == NULL)
                {
                    /* No extension. */
        
#ifdef USE_IMLIB
                    /* Is extension an image? */
                    if(ISEXTIMAGE(tmp_path))
                    {
                        strcpy(cmd, "%imgview ");
                        strncat(cmd, edw->file[file_num]->name,
                            PATH_MAX + NAME_MAX - strlen(cmd) - 1
                        );
                    }
                    else
                    {
                        /* Format cmd to use regular internal viewer. */
                        strcpy(cmd, "%view ");
                        strncat(cmd, edw->file[file_num]->name,
                            PATH_MAX + NAME_MAX - strlen(cmd) - 1
                        );
                    }
#else /* USE_IMLIB */
                    /* Format cmd to use regular internal viewer. */
                    strcpy(cmd, "%view ");
                    strncat(cmd, edw->file[file_num]->name,
                        PATH_MAX + NAME_MAX - strlen(cmd) - 1
                    );
#endif /* USE_IMLIB */
                }
                else
                {
                    /* It has an extension, set the associated command. */
                    strncpy(cmd, strptr, PATH_MAX + NAME_MAX + 256);
                
                    /* Substitute tmp_path into cmd properly. */
                    FAssSubCommand(
			cmd,
			PrefixPaths(
			    edw->loc_prompt.buf,
			    edw->file[file_num]->name
			)
		    );
                }
                cmd[PATH_MAX + NAME_MAX + 256 - 1] = '\0';
            }   
            else
            {
                /* Has no extension, check if it is set executeable. */
                if(ISPATHEXECUTABLE(tmp_path))   
                {
                    /* It is executeable, set command to be tmp_path. */
                    strcpy(cmd, tmp_path);
                }
                else
                {
                    /* Not executeable and no extension, use internal viewer. */
                    strcpy(cmd, "%view ");
                    strncat(cmd, edw->file[file_num]->name,
                        PATH_MAX + NAME_MAX - strlen(cmd) - 1
                    );
                }
                cmd[PATH_MAX + NAME_MAX + 256 - 1] = '\0';
            }
                    
            /* The cmd at this point should be defined properly. */
            /* ********************************************************* */
                      

            /* Update the cmd_prompt. */
	    PromptSetS(&edw->cmd_prompt, cmd);
                

            /* Do we want to execute the command? */
            if(!execute)
		continue;


            /* Execute command. */
            EDWDoExecCmdPrompt(edw);
            executions++;
        }


        return(0);
}



/*
 *	Endeavour Browser menu callback.
 */
int EDWMenuCB(void *client_data, int id_code)
{
	int i, n;
        char stringa[PATH_MAX + NAME_MAX + 128]; 
        endeavour_win_struct *edw;
	int status = 0;


        /* Error checks. */
        if(!IDC() ||
           (client_data == NULL)
        )
            return(-1);


        /* ************************************************************** */
        
        /* Search for valid img_browser by client_data pointer. */
        for(i = 0; i < total_endeavour_wins; i++)
        {
            if(endeavour_win[i] == (endeavour_win_struct *)client_data)
                break;
        }
        /* Could not find match? */
        if(i >= total_endeavour_wins)
            return(-2);

        /* Is initialized? */
        if(!EDWIsInit(i))
            return(-1);
        /* Is mapped? */  
        if(endeavour_win[i]->map_state == 0)
            return(-1);

        /* Get pointer to edw. */
        edw = (endeavour_win_struct *)endeavour_win[i];


        /* Handle menu function. */
        switch(id_code)      
        {
          /* ************************************************************ */
          case EDW_MC_NONE:
            break;

          case EDW_MC_DIR_EXPAND_TOGGLE:
	    if(edw->total_sel_dirs <= 0) break;
	    n = edw->sel_dir[0];
	    if((n < 0) || (n >= edw->total_dirs)) break;
	    if(edw->dir[n] == NULL) break;

	    if(edw->dir[n]->expanded)
		EDWColapseDir(edw, n);
	    else
		EDWExpandDir(edw, n);
	    EDWDraw(edw, DRAW_AMOUNT_COMPLETE);
            break;

	  case EDW_MC_FILE_EXECUTE:
	    EDWExecuteSelectedFiles(edw, True);
	    break;

          /* ******************************************************** */
	  case EDW_MC_COPY:
	    EDWDoCopy(edw);
            break;

          case EDW_MC_MOVE:
            EDWDoMove(edw);
            break;

          case EDW_MC_RENAME:
            if(edw->total_sel_dirs > 0)
	    {
		EDWDoMapCVPromptDir(edw, edw->sel_dir[0],
                    CV_PROMPT_MODE_RENAME);
	    }
	    else if(edw->total_sel_files > 0)
            {
                EDWDoMapCVPromptFile(edw, edw->sel_file[0],
                    CV_PROMPT_MODE_RENAME);
            }
            break;

          case EDW_MC_RECYCLE:
            EDWDoRecycle(edw);
            break;

          case EDW_MC_PROPERTIES:
            EDWDoEditProperties(edw);
            break;

          /* ********************************************************* */
	  case EDW_MC_EDITOR:
            /* edit selected files. */
            for(i = 0; i < edw->total_sel_files; i++)
            {
                n = edw->sel_file[i];
                if((n < 0) || (n >= edw->total_files)) continue;
                if(edw->file[n] == NULL) continue;

                /* Run text editor. */
                FileOpRunTextEditor(edw->file[n]->name);
            }
	    break;

          /* ********************************************************* */
          default:
            sprintf(stringa,
"Internal non-critical error:\n\n\
    Menu function code %i not supported.\n\n",
                id_code
            );
            printdw(&err_dw, stringa);  
            break;
	}


	return(status);
}
