/*
                 Endeavour Browser: Selecting List Items

	Functions:

        int EDW_ISDIRSELECTED(endeavour_win_struct *edw, int dir_num)
        int EDW_ISFILESELECTED(endeavour_win_struct *edw, int file_num)

        int EDW_SELECTDIR(endeavour_win_struct *edw, int dir_num)
        int EDW_SELECTFILE(endeavour_win_struct *edw, int file_num)

        void EDW_UNSELECTDIR(endeavour_win_struct *edw, int dir_num)
        void EDW_UNSELECTFILE(endeavour_win_struct *edw, int file_num)

        void EDW_UNSELECTALLDIRS(endeavour_win_struct *edw)
        void EDW_UNSELECTALLFILES(endeavour_win_struct *edw)


	int EDWDoSelectDir(endeavour_win_struct *edw, int dir_num)

	---

 */

#include "endeavour.h"
#include "edwdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


/*
 *	Returns true if dir_num is selected on edw.
 */
int EDW_ISDIRSELECTED(endeavour_win_struct *edw, int dir_num)
{
	int i;

        if(edw == NULL)
            return(0);
            
        for(i = 0; i < (int)edw->total_sel_dirs; i++)
        {
            if(edw->sel_dir[i] == dir_num)
                return(1);
        }

        return(0);
}


/*
 *      Returns true if file_num is selected on edw.
 */
int EDW_ISFILESELECTED(endeavour_win_struct *edw, int file_num)
{
        int i;

        if(edw == NULL)
            return(0);  

        for(i = 0; i < (int)edw->total_sel_files; i++)
        {
            if(edw->sel_file[i] == file_num)
                return(1);
        }

        return(0);
}


/*
 *	Selects dir dir_num on edw.
 */
int EDW_SELECTDIR(endeavour_win_struct *edw, int dir_num)
{
        int i;


        if(edw == NULL)
            return(-1);

        if(edw->total_dirs < 0) 
            edw->total_dirs = 0;
        if(edw->total_sel_dirs < 0)
            edw->total_sel_dirs = 0;

	if((dir_num < 0) || (dir_num >= edw->total_dirs))
	    return(-2);


        /* Check if is already 'selected' */
        for(i = 0; i < edw->total_sel_dirs; i++)
        {
            if(edw->sel_dir[i] == dir_num)
                return(-2);
        }

        /* Allocate more selections. */
        edw->total_sel_dirs++;

        edw->sel_dir = (int *)realloc(edw->sel_dir,
            edw->total_sel_dirs * sizeof(int)
        );
        if(edw->sel_dir == NULL)
        {
            edw->total_sel_dirs = 0;
            return(-1);
        }

        /* Set new value. */
        edw->sel_dir[edw->total_sel_dirs - 1] = dir_num;

        return(0);
}


/*
 *      Selects file file_num on edw.
 */
int EDW_SELECTFILE(endeavour_win_struct *edw, int file_num)
{
        int i;


        if(edw == NULL)
            return(-1);

        if(edw->total_files < 0)
            edw->total_files = 0;
        if(edw->total_sel_files < 0)
            edw->total_sel_files = 0;

        if((file_num < 0) || (file_num >= edw->total_files))
            return(-2);


        /* Check if is already 'selected' */
        for(i = 0; i < (int)edw->total_sel_files; i++)
        {
            if(edw->sel_file[i] == file_num)
                return(-2);
        }

        /* Allocate more selections. */
        edw->total_sel_files++;

        edw->sel_file = (int *)realloc(edw->sel_file,
            (int)edw->total_sel_files * sizeof(int)  
        );
        if(edw->sel_file == NULL)
        {
            edw->total_sel_files = 0;
            return(-1);
        }

        /* Set new value. */
        edw->sel_file[(int)edw->total_sel_files - 1] = file_num;

        return(0);
}

/*
 *	Unselects dir dir_num on edw.
 */
void EDW_UNSELECTDIR(endeavour_win_struct *edw, int dir_num)
{
        int i;
        
        if(edw == NULL)
            return;
            
        if((dir_num < 0) || (dir_num >= (int)edw->total_dirs))
            return;
            
        for(i = 0; i < (int)edw->total_sel_dirs; i++)
        {
            if(edw->sel_dir[i] == dir_num)
                edw->sel_dir[i] = -1;
        }

        return;
}

/*
 *      Unselects file file_num on edw.
 */
void EDW_UNSELECTFILE(endeavour_win_struct *edw, int file_num)
{
        int i;

        if(edw == NULL)
            return;

        if((file_num < 0) || (file_num >= (int)edw->total_files))
            return;

        for(i = 0; i < (int)edw->total_sel_files; i++)
        {
            if(edw->sel_file[i] == file_num)
                edw->sel_file[i] = -1;
        }
         
        return;
}


/*
 *      Macro which unselects all directories on edw's dir list.
 *
 *      Note that technically only one or less directory may be selected at
 *      any given time.
 */
void EDW_UNSELECTALLDIRS(endeavour_win_struct *edw)
{
        if(edw == NULL)
            return;

        free(edw->sel_dir);
        edw->sel_dir = NULL;

        edw->total_sel_dirs = 0;

        return;
}

/*
 *      Macro which unselects all files on edw's file list.
 */
void EDW_UNSELECTALLFILES(endeavour_win_struct *edw)
{
        if(edw == NULL)
            return;
 
        free(edw->sel_file);
        edw->sel_file = NULL;
  
        edw->total_sel_files = 0;

        return;
}

/*
 *	Procedure to select a dir on edw by number.
 */
int EDWDoSelectDir(endeavour_win_struct *edw, int dir_num)
{
	int i, n, got_dev;
	char stringa[PATH_MAX + NAME_MAX + 128];
	int status = 0;


        if(edw == NULL)
            return(-1);

	/* Need to unselect all files and dirs first. */
	EDW_UNSELECTALLDIRS(edw);
        EDW_UNSELECTALLFILES(edw);


	/* Select the dir. */
	EDW_SELECTDIR(edw, dir_num);

	if(edw->total_sel_dirs > 0)
	{
	    i = edw->sel_dir[0];
	    if((i >= 0) && (i < edw->total_dirs))
	    {
		if(edw->dir[i] != NULL)
		{
		    /* Update the file listing. */
	            EDWRefreshFileList(edw, edw->dir[i]->fullname);

		    /* Reset file list scroll positions. */
                    edw->file_win_sb.x_win_pos = 0;
                    edw->file_win_sb.y_win_pos = 0;

		    /* Update the location prompt buffer. */
		    if(edw->loc_prompt.buf != NULL)
		    {
			strncpy(edw->loc_prompt.buf,
			    edw->dir[i]->fullname,
			    edw->loc_prompt.buf_len
			);
			edw->loc_prompt.buf_pos =
			    strlen(edw->loc_prompt.buf);
		    }


                    /* Update status message, free it first. */
		    free(edw->status_mesg);
                    edw->status_mesg = NULL;  

		    /* Is it a device? */
		    got_dev = 0;
		    if(edw->dir[i]->fullname != NULL)
		    {
			for(n = 0; n < total_devrecs; n++)
			{
			    if(!DevRecIsAllocated(n))
                                continue; 

			    if(strcmp(edw->dir[i]->fullname,
				devrec[n]->path)
			    )
				continue;

                            if(devrec[n]->is_mounted)
                            {
                                sprintf(stringa,
 "Device `%s' Total: %ld  Used: %ld  Available: %ld  Ratio: %.0f%%",
                                    devrec[n]->label,
                                    devrec[n]->capacity_total,
                                    devrec[n]->capacity_used,
                                    devrec[n]->capacity_available,
                                    devrec[n]->capacity_percent
                                );
                            }
                            else
                            {
                                sprintf(stringa,
 "Device `%s' not mounted (right click to mount on directory `%s').",
                                    devrec[n]->label,
                                    devrec[n]->path 
                                );
                            }

                            edw->status_mesg = StringCopyAlloc(stringa); 

			    got_dev = 1;
			    break;
			}
		    }

		    /* Regular dir. */
		    if((edw->dir[i]->name != NULL) &&
                       !got_dev
                    )
		    {
			if(edw->dir[i]->expandable)
			{
                            sprintf(stringa,
  "Directory `%s' selected (double click to %s that directory)",
			        edw->dir[i]->name,
			        (edw->dir[i]->expanded) ?
				    "colapse" : "expand"
			    );
			}
			else
                        {
                            sprintf(stringa,
                                "Directory `%s' selected",
				edw->dir[i]->name
                            );
                        }
			edw->status_mesg = StringCopyAlloc(stringa);
		    }
		}
	    }
	}
	else
	{
	    status = -1;
	}

	return(status);
}
