/*
                     File Associations Handling

	Functions:

	int ISEXTIMAGE(char *filename)

	void FAssSubCommand(char *cmd, char *arg)

	void FAssLoadDefaults()
	int FAssAdd(char *ext, char *cmd)
	void FAssDeleteAll()

	char *FAssExtParse(char *string)
	char *FAssGetCmdByExt(char *ext)

 */

#include "../include/string.h"
#include "fass.h"




/*
 *	Checks if the file's name has an extension that is a known
 *	(known by this program) to be an image format.
 */
int ISEXTIMAGE(char *filename)
{
	int i;
	char ext[MAX_EXT_LEN];
	char ext_test[MAX_EXT_LEN];
	const char *cstrptr;
	const char *known_image_ext[] = FASS_KNOWN_IMAGE_EXT;

	/* Parse extension from filename. */
	cstrptr = (const char *)FAssExtParse(filename);
	if(cstrptr == NULL)	/* No extension? */
	    return(0);

	strncpy(ext, cstrptr, MAX_EXT_LEN);
	ext[MAX_EXT_LEN - 1] = '\0';

	strtoupper(ext);

	for(i = 0; known_image_ext[i] != NULL; i++)
	{
	    strncpy(ext_test, known_image_ext[i], MAX_EXT_LEN);
	    ext_test[MAX_EXT_LEN - 1] = '\0';
	    strtoupper(ext_test);

	    if(!strcmp(ext, ext_test))
		return(1);
	}


	return(0);
}



/*
 *	Substitutes any occurace of "%s" in string cmd with
 *	the string contained in arg.
 */
void FAssSubCommand(char *cmd, char *arg)
{
	int len1, len2;
        char *tmp_str;
        char *strptr, *strptr2;


	/* Error checks. */
	if(cmd == NULL)
	    return;

	if(arg == NULL)
	    arg = "(null)";

	/* ******************************************************** */

        /* Get lengths. */
        len1 = strlen(cmd);
        len2 = strlen(arg);


        /* Allocate memory for tmp_str. */
/* Need to compensate for multiple subs!!! */
        tmp_str = (char *)calloc(1, (len1 + len2 + 2) * sizeof(char));
        if(tmp_str == NULL)
            return;

        strcpy(tmp_str, cmd);


	/* Begin substitution. */
	while(1)
        {
	    strptr = strstr(tmp_str, "%s");
            if(strptr == NULL)
		break;

            strcpy(strptr, arg);

            strptr2 = strstr(cmd, "%s") + 2;  
            strcpy(strptr + len2, strptr2);
	}


	/* Copy tmp_str back to cmd which must have enough storage. */
        strcpy(cmd, tmp_str);


  	/* Free allocated tmp_str. */
        free(tmp_str); tmp_str = NULL;


	return;
}



/*
 *	Loads default associations.
 */
void FAssLoadDefaults()
{
	int status;


	/* Clear all file associations. */
	FAssDeleteAll();


        /* C sources. */
        status = FAssAdd(
            "c cpp cc h",
"nxterm -geometry 80x30 -fn 8x16 -sl 512 -fg \"#ffffff\" \
-bg \"#000000\" -e pico %s"
        );
        if(status < 0)
            return;

        /* Wave or voc audio. */
        status = FAssAdd(
            "voc wav wave riff", 
            "yiff %s"
        );
        if(status < 0)
            return;

        /* MP3 audio. */
        status = FAssAdd(
            "mp3",
            "mxaudio %s"
        );
        if(status < 0)   
            return;


        /* Multimedia. */
        status = FAssAdd(
            "mpeg mpg mov avi",
            "xanim -q %s"
        );
        if(status < 0)
            return;



	return;
}



int FAssAdd(char *ext, char *cmd)
{
	int i, len;
	int new_entry_num;


	if(total_file_asses < 0)
	    total_file_asses = 0;


	/* Look for NULL entry. */
	for(i = 0; i < (int)total_file_asses; i++)
	{
	    if(file_ass[i] == NULL)
		break;
	}
	if(i < (int)total_file_asses)
	{
	    /* Found one, allocate. */
	    file_ass[i] = (file_ass_struct *)calloc(1,
		sizeof(file_ass_struct));
	    if(file_ass[i] == NULL)
		return(-1);

	    new_entry_num = i;
	}
	else
	{
	    new_entry_num = total_file_asses;
            total_file_asses++;

	    file_ass = (file_ass_struct **)realloc(file_ass,
		total_file_asses * sizeof(file_ass_struct *));
	    if(file_ass == NULL)
	    {
		total_file_asses = 0;
		return(-1);
	    }

	    file_ass[new_entry_num] = (file_ass_struct *)calloc(1,
		sizeof(file_ass_struct));
	    if(file_ass[new_entry_num] == NULL)
	    {
		return(-1);
	    }
	}


	/* Add extension. */
	if(ext != NULL)
	{
	    /* Rmember to add one space just before and after! */
	    len = (strlen(ext) + 3) * sizeof(char);

	    file_ass[new_entry_num]->ext = (char *)calloc(1, len);
	    if(file_ass[new_entry_num]->ext == NULL)
		return(-1);
	    strcpy(file_ass[new_entry_num]->ext + 1, ext);

	    file_ass[new_entry_num]->ext[0] = ' ';
	    file_ass[new_entry_num]->ext[len - 2] = ' ';
	}

        /* Add command. */
        if(cmd != NULL)
        {
            len = (strlen(cmd) + 1) * sizeof(char);
         
            file_ass[new_entry_num]->cmd = (char *)calloc(1, len);               
            if(file_ass[new_entry_num]->cmd == NULL)
                return(-1);
            strcpy(file_ass[new_entry_num]->cmd, cmd);
        }




	return(new_entry_num);
}



void FAssDeleteAll()
{
	int i;


	for(i = 0; i < (int)total_file_asses; i++)
	{
	    if(file_ass[i] == NULL) continue;

	    free(file_ass[i]->ext);
	    free(file_ass[i]->cmd);
	    free(file_ass[i]); file_ass[i] = NULL;
	}

	free(file_ass); file_ass = NULL;
	total_file_asses = 0;


	return;
}



/*
 *	Returns a statically allocated string containing the extension
 *	parsed from string.
 *
 *	Can return NULL on error or if no extension could be found.
 */
char *FAssExtParse(char *string)
{
	char *fn_start_ptr;
        char *strptr;
	static char rtn_str[MAX_EXT_LEN];


	if(string == NULL)
	    return(NULL);


	/* Skip directory delimintations (if any). */
	strptr = strrchr(string, '/');	/* Search backwards. */
	if(strptr == NULL)
	    fn_start_ptr = string;
	else
	    fn_start_ptr = strptr + 1;


	/* Search for the extension deliminator character (if any). */
	strptr = strrchr(string, EXT_DELIMITER_CHAR);	/* Search backwards. */
	if(strptr == NULL)
	{
	    /* Return NULL on no extension. */
	    return(NULL);
	}
	else
	{
	    /* Got extension, copy it to the return string. */
	    strncpy(rtn_str, strptr + 1, MAX_EXT_LEN);
	    rtn_str[MAX_EXT_LEN - 1] = '\0';

	    return(rtn_str);
	}
}



char *FAssGetCmdByExt(char *ext)
{
	int i, len;
	char stringa[MAX_EXT_FIELD_LEN];
	char lext[MAX_EXT_LEN];


	/* Error checks. */
	if(ext == NULL)
	    return(NULL);

	/* Copy ext to local. */
	strncpy(&(lext[1]), ext, MAX_EXT_LEN - 1);
	lext[MAX_EXT_LEN - 1] = '\0';

	lext[0] = EXT_FIELD_DELIMITER_CHAR;
	len = strlen(lext);
	lext[len] = EXT_FIELD_DELIMITER_CHAR;
	lext[len + 1] = '\0';



	for(i = 0; i < (int)total_file_asses; i++)
	{
	    if(file_ass[i] == NULL) continue;
	    if(file_ass[i]->ext == NULL) continue;

	    strncpy(
		&(stringa[1]),
		file_ass[i]->ext,
		MAX_EXT_FIELD_LEN - 1
	    );
	    stringa[MAX_EXT_FIELD_LEN - 1] = '\0';

	    stringa[0] = EXT_FIELD_DELIMITER_CHAR;
	    len = strlen(stringa);
	    stringa[len] = EXT_FIELD_DELIMITER_CHAR;
	    stringa[len + 1] = '\0';

	    strtoupper(stringa);
	    strtoupper(lext);

	    if(strstr(stringa, lext) != NULL)
		return(file_ass[i]->cmd);
	}


	return(NULL);
}
