/*
                          File Operations

	Functions:

	void STRINGSUBTARGET(char *string, char *sub)
	void STRINGSUBSOURCE(char *string, char *sub)
	void STRINGSUBARG(char *string, char *sub)
	int DESTEXISTS(char *tar_path, char *src_path)
	int ISCMDREMOVE(char *cmd)

	int FileOpMoveFile(char *tar_path, char *src_path)
	int FileOpMoveDir(char *tar_path, char *src_path)

	int FileOpCopyFile(char *tar_path, char *src_path)
	int FileOpCopyDir(char *tar_path, char *src_path)

	int FileOpRenameFile(char *tar_path, char *src_path)
	int FileOpRenameDir(char *tar_path, char *src_path)

	int FileOpChmodFile(char *tar_path, char *mode_str)
	int FileOpChmodDir(char *tar_path, char *mode_str)

	int FileOpChownOwner(char *tar_path, char *owner)
	int FileOpChownGroup(char *tar_path, char *group)

        int FileOpTouchFile(char *tar_path)
        int FileOpTouchDir(char *tar_path)

	int FileOpRunTextEditor(char *tar_path)

*/

#include "endeavour.h"


#ifndef CMD_MAX
    #define CMD_MAX	2048
#endif


/*    List of remove or delete commands that should not be allowed 
 *    to execute if write-protect is on.
 */
char *remove_cmds[] = {
        "rm", "remove", "del", "delete", "erase",
        "rmdir", NULL
};



/* ****************************************************************** */

/*
 *	Substitute sub in string replacing all occurances of "%t".
 */
void STRINGSUBTARGET(char *string, char *sub)
{
	int len1, len2;
        char *tmp_str;
        char *strptr, *strptr2;


        /* Error checks. */
        if(string == NULL)
            return;
            
        if(sub == NULL)
            sub = "(null)";
        

        /* Get lengths. */
        len1 = strlen(string);
        len2 = strlen(sub);

        /* Begin substitution. */
        tmp_str = (char *)calloc(1, (len1 + len2 + 2) * sizeof(char));
        if(tmp_str == NULL)
            return;
            
        strcpy(tmp_str, string);

        strptr = strstr(tmp_str, "%t");
        if(strptr == NULL)
        {
            free(tmp_str); tmp_str = NULL;
            return;
        }

        strcpy(strptr, sub);
        
        strptr2 = strstr(string, "%t") + 2;  
        strcpy(strptr + len2, strptr2);
        
        
        strcpy(string, tmp_str);        /* String must have enough storage. */

  
        free(tmp_str); tmp_str = NULL;


        return;
}



/*
 *      Substitute sub in string replacing all occurances of "%s".
 */
void STRINGSUBSOURCE(char *string, char *sub)
{
        int len1, len2;
        char *tmp_str;
        char *strptr, *strptr2;


        /* Error checks. */
        if(string == NULL)
            return;
            
        if(sub == NULL)
            sub = "(null)";
        

        /* Get lengths. */
        len1 = strlen(string);
        len2 = strlen(sub);

        /* Begin substitution. */
        tmp_str = (char *)calloc(1, (len1 + len2 + 2) * sizeof(char));
        if(tmp_str == NULL)
            return;
            
        strcpy(tmp_str, string);

        strptr = strstr(tmp_str, "%s");
        if(strptr == NULL)
        {
            free(tmp_str); tmp_str = NULL;
            return;
        }

        strcpy(strptr, sub);
        
        strptr2 = strstr(string, "%s") + 2;  
        strcpy(strptr + len2, strptr2);
        
        
        strcpy(string, tmp_str);        /* String must have enough storage. */

  
        free(tmp_str); tmp_str = NULL;


        return;
}



/*
 *      Substitute sub in string replacing all occurances of "%a".
 */
void STRINGSUBARG(char *string, char *sub)
{
	int len1, len2;
        char *tmp_str;
        char *strptr, *strptr2;

        
        /* Error checks. */
        if(string == NULL)
            return;
         
        if(sub == NULL)
            sub = "(null)";
         

        /* Get lengths. */  
        len1 = strlen(string);
        len2 = strlen(sub);

        /* Begin substitution. */
        tmp_str = (char *)calloc(1, (len1 + len2 + 2) * sizeof(char));
        if(tmp_str == NULL)
            return;

        strcpy(tmp_str, string);

        strptr = strstr(tmp_str, "%a");
        if(strptr == NULL)
        {
            free(tmp_str); tmp_str = NULL;
            return;
        }

        strcpy(strptr, sub);

        strptr2 = strstr(string, "%a") + 2;
        strcpy(strptr + len2, strptr2);

        strcpy(string, tmp_str);        /* String must have enough storage. */


        free(tmp_str); tmp_str = NULL;

        
        return;
}



/*
 *	Returns 1 if:
 *
 *	if src_path is a directory, then does tar_path's destination name
 *	exist in src_path.
 *
 *	if src_path is a file or other object, then is tar_path the same
 *	as src_path.
 */
int DESTEXISTS(char *tar_path, char *src_path)
{
	char tmp_path[PATH_MAX + NAME_MAX];
	char tar_name[PATH_MAX + NAME_MAX];
	char *strptr1;
	struct stat stat_buf;


	/* Error checks. */
	if((tar_path == NULL) ||
           (src_path == NULL)
	)
	    return(0);


	/* Check if tar_path is a directory. */
	if(ISPATHDIR(tar_path))
	{
	    /* Get tar_name (name only, not absolute path) of src_path. */
	    strncpy(tar_name, src_path, PATH_MAX + NAME_MAX);
	    tar_name[PATH_MAX + NAME_MAX - 1] = '\0';
	    StripAbsolutePath(tar_name);

	    strptr1 = PrefixPaths(tar_path, tar_name);
	    if(strptr1 == NULL) return(0);


	    /* Test tmp_path, the target path with the source name postpended. */
	    strncpy(tmp_path, strptr1, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    if(lstat(tmp_path, &stat_buf))
		return(0);	/* Source does NOT exist in target dir. */
            else
                return(1);
	}
	else
	{
	    /* Target is not a directory, so just check if it exists. */
	    if(lstat(tar_path, &stat_buf))
		return(0);	/* Target does NOT exist. */
	    else
		return(1);
	}


	return(0);
}



/*
 *	Checks if the given command string is a remove command.
 *	Returns 1 if it is, 0 if it is not.
 */
int ISCMDREMOVE(char *cmd)
{
	char *strptr, *cmd_only_str;

	int x, status;


	/* Error checks. */
	if(cmd == NULL) return(0);


	/* Skip leading spaces. */
	for(x = 0; cmd[x] == ' '; x++) continue;

	/* Copy cmd to cmd_only_str. */
	cmd_only_str = StringCopyAlloc(&(cmd[x]));
	if(cmd_only_str == NULL) return(0);

	/* Seek end. */
	strptr = strchr(cmd_only_str, ' ');
	if(strptr != NULL)
	{
	    *strptr = '\0';
	}


	/* Go through remove_cmds. */
	status = 0;
	for(x = 0; remove_cmds[x] != NULL; x++)
	{
	    if(!strcmp(cmd_only_str, remove_cmds[x]))
		status = 1;
	}

	/* Free cmd_only_str. */
	free(cmd_only_str);
	cmd_only_str = NULL;


	return(status);
}



/* ************************************************************************ */

/*
 *	Runs the move file command operation on the given paths.
 */
int FileOpMoveFile(char *tar_path, char *src_path)
{
	char cmd[CMD_MAX];


	/* Error checks. */
	if(command.move_file == NULL)
	    return(-1);


	/* Remove stdout spool file. */
	SFRemoveProgStdout();

	/* Parse and move file (use external program). */
	strncpy(cmd, command.move_file, CMD_MAX);
	cmd[CMD_MAX - 1] = '\0';
	STRINGSUBTARGET(cmd, tar_path);
	STRINGSUBSOURCE(cmd, src_path);
        cmd[CMD_MAX - 1] = '\0';

	/* Execute the command. */
	ExecBAOE(cmd, fname.prog_output, fname.prog_output);

	/* Show errors if any. */
	SFShowOutputOnDialog(&err_dw);


	return(0);
}



/* 
 *      Runs the move dir command operation on the given paths.
 */
int FileOpMoveDir(char *tar_path, char *src_path)
{
        char cmd[CMD_MAX];


        /* Error checks. */
        if(command.move_dir == NULL)
            return(-1);


        /* Remove stdout spool file. */
        SFRemoveProgStdout();

        /* Parse and move dir (use external program). */
        strncpy(cmd, command.move_dir, CMD_MAX);
        cmd[CMD_MAX - 1] = '\0';
        STRINGSUBTARGET(cmd, tar_path);
        STRINGSUBSOURCE(cmd, src_path);
        cmd[CMD_MAX - 1] = '\0';

        /* Execute the command. */
        ExecBAOE(cmd, fname.prog_output, fname.prog_output);

        /* Show errors if any. */
        SFShowOutputOnDialog(&err_dw);

  
        return(0);
}



/*
 *	Runs the copy file command operation on the given paths.
 */
int FileOpCopyFile(char *tar_path, char *src_path)
{
        char cmd[CMD_MAX];


        /* Error checks. */
        if(command.copy_file == NULL)
            return(-1);


        /* Remove stdout spool file. */
        SFRemoveProgStdout();

	/* Parse and copy file (use external program). */
        strncpy(cmd, command.copy_file, CMD_MAX);
        cmd[CMD_MAX - 1] = '\0';
        STRINGSUBTARGET(cmd, tar_path);
        STRINGSUBSOURCE(cmd, src_path);
        cmd[CMD_MAX - 1] = '\0';

        /* Execute the command. */
        ExecBAOE(cmd, fname.prog_output, fname.prog_output);

        /* Show errors if any. */
        SFShowOutputOnDialog(&err_dw);


        return(0);
}



/*
 *	Runs the copy dir command operation on the given paths.
 */
int FileOpCopyDir(char *tar_path, char *src_path)
{
        char cmd[CMD_MAX];


        /* Error checks. */
        if(command.copy_dir == NULL)
            return(-1);


        /* Remove stdout spool file. */
        SFRemoveProgStdout();

	/* Parse and copy dir (use external program). */
        strncpy(cmd, command.copy_dir, CMD_MAX);
        cmd[CMD_MAX - 1] = '\0';
        STRINGSUBTARGET(cmd, tar_path);
        STRINGSUBSOURCE(cmd, src_path);
        cmd[CMD_MAX - 1] = '\0';

        /* Execute the command. */
        ExecBAOE(cmd, fname.prog_output, fname.prog_output);

        /* Show errors if any. */
        SFShowOutputOnDialog(&err_dw);


        return(0);
}




int FileOpRenameFile(char *tar_path, char *src_path)
{
        char cmd[CMD_MAX];


        /* Error checks. */
        if(command.rename_file == NULL)
            return(-1);


        /* Remove stdout spool file. */
        SFRemoveProgStdout();

	/* Parse and rename file (use external program). */
        strncpy(cmd, command.rename_file, CMD_MAX);
        cmd[CMD_MAX - 1] = '\0';
        STRINGSUBTARGET(cmd, tar_path);
        STRINGSUBSOURCE(cmd, src_path);
        cmd[CMD_MAX - 1] = '\0';

        /* Execute the command. */
        ExecBAOE(cmd, fname.prog_output, fname.prog_output);

        /* Show errors if any. */
        SFShowOutputOnDialog(&err_dw);

        
        return(0);
}



int FileOpRenameDir(char *tar_path, char *src_path)
{
        char cmd[CMD_MAX];


        /* Error checks. */
        if(command.rename_dir == NULL)
            return(-1);


        /* Remove stdout spool file. */
        SFRemoveProgStdout();

        /* Parse and rename dir (use external program). */
        strncpy(cmd, command.rename_dir, CMD_MAX);
        cmd[CMD_MAX - 1] = '\0';
        STRINGSUBTARGET(cmd, tar_path);
        STRINGSUBSOURCE(cmd, src_path);
        cmd[CMD_MAX - 1] = '\0';

        /* Execute the command. */
        ExecBAOE(cmd, fname.prog_output, fname.prog_output);

        /* Show errors if any. */
        SFShowOutputOnDialog(&err_dw);


        return(0);
}



/*
 *	Chmod (change permissions):
 */
int FileOpChmodFile(char *tar_path, char *mode_str)
{
	int status;
	mode_t mode;
	char lmode_str[10];


	if( (tar_path == NULL) ||
            (mode_str == NULL)
	)
	    return(-1);


	/* Copy mode_str to lmode_str. */
	memset(lmode_str, 0x00, 10 * sizeof(char));
	strncpy(lmode_str, mode_str, 10);
	lmode_str[10 - 1] = '\0';


	mode = 0;

	mode |= (lmode_str[0] == 'r') ? S_IRUSR : 0;
	mode |= (lmode_str[1] == 'w') ? S_IWUSR : 0;
	mode |= (lmode_str[2] == 'x') ? S_IXUSR : 0;
	if(lmode_str[2] == 's')		/* Set UID. */
	    mode |= S_ISUID;

        mode |= (lmode_str[3] == 'r') ? S_IRGRP : 0;
        mode |= (lmode_str[4] == 'w') ? S_IWGRP : 0;
        mode |= (lmode_str[5] == 'x') ? S_IXGRP : 0;
        if(lmode_str[5] == 's')         /* Set GID. */
            mode |= S_ISGID;

        mode |= (lmode_str[6] == 'r') ? S_IROTH : 0;
        mode |= (lmode_str[7] == 'w') ? S_IWOTH : 0;
        mode |= (lmode_str[8] == 'x') ? S_IXOTH : 0;
        if(lmode_str[8] == 't')         /* Set GID. */
            mode |= S_ISVTX;

	status = chmod(tar_path, mode);


	return(status);
}



int FileOpChmodDir(char *tar_path, char *mode_str)
{
	/* Same as  FileOpChmodFile. */
	return(FileOpChmodFile(tar_path, mode_str));
}



/* *********************************************************************
 *
 */
int FileOpChownOwner(char *tar_path, char *owner)   
{
        int status;
	struct stat stat_buf;
	uid_t uid;
	gid_t gid;


	if(owner == NULL) return(-1);

	/* Get lstats. */
	if(lstat(tar_path, &stat_buf) == -1)
	    return(-1);

	uid = IDGetUIDbyName(owner);
	gid = stat_buf.st_gid;


        /* Chown using C call. */
        status = chown(tar_path, uid, gid);

	return(status);
}


int FileOpChownGroup(char *tar_path, char *group)  
{
	int status;
        struct stat stat_buf;
        uid_t uid;
        gid_t gid;


	if(group == NULL) return(-1);

        /* Get lstats. */
        if(lstat(tar_path, &stat_buf) == -1)
            return(-1);

	uid = stat_buf.st_uid;
	gid = IDGetGIDbyGroup(group);


	/* Chown using C call. */
	status = chown(tar_path, uid, gid);

	return(status);
}



/* *********************************************************************
 *
 */
int FileOpTouchFile(char *tar_path)
{
	struct stat stat_buf;
        char cmd[CMD_MAX];


	/* Error checks. */
	if((tar_path == NULL) ||
           (command.touch_file == NULL)
	)
	    return(-1);

	if(lstat(tar_path, &stat_buf) == -1)
	    return(-1);


	/* Remove stdout spool file. */
	SFRemoveProgStdout();

	/* Substitude command line. */
	strncpy(cmd, command.touch_file, CMD_MAX);
	cmd[CMD_MAX - 1] = '\0';
	STRINGSUBTARGET(cmd, tar_path);
	cmd[CMD_MAX - 1] = '\0';

        /* Execute the command. */
        ExecBAOE(cmd, fname.prog_output, fname.prog_output);

        /* Show errors if any. */
        SFShowOutputOnDialog(&err_dw);


	return(0);
}



int FileOpTouchDir(char *tar_path)  
{
	return(FileOpTouchFile(tar_path));
}


/*
 *	Executes text editor:
 */
int FileOpRunTextEditor(char *tar_path)
{
        struct stat stat_buf;
        char cmd[CMD_MAX];
	pid_t p;


	/* Error checks. */
        if((tar_path == NULL) ||
           (command.text_editor == NULL)
        )   
            return(-1);

	/*   Check if target object exists (hope calling function
         *   changed working directory for us).
	 */
        if(stat(tar_path, &stat_buf))
            return(-1);


        /* Substitude command line. */
        strncpy(cmd, command.text_editor, CMD_MAX);
        cmd[CMD_MAX - 1] = '\0';
        STRINGSUBTARGET(cmd, tar_path);
        cmd[CMD_MAX - 1] = '\0';

        /* Execute text editor in background. */
        p = ExecAOE(cmd, fname.prog_output, fname.prog_output);

	/* Record process history. */
	ProcHistAdd(p, time(NULL));


	return(0);
}
