/*
                             ID Record Management

	Functions:

	int ID_ISSTRINGNUM(char *string) 

	char *IDGetNameByUID(uid_t uid)
	char *IDGetGroupByGID(gid_t gid)
	uid_t IDGetUIDbyName(char *name)
	gid_t IDGetGIDbyGroup(char *group)

	int IDAllocate()
	int IDLoadFromFile(char *filename)
	void IDDeleteAll()


*/

#include "endeavour.h"



/*
 *	Returns 1 if the first char in string is a number.
 */
int ID_ISSTRINGNUM(char *string)
{
        if(string == NULL)
	    return(0);

        return(isdigit((int)string[0]));
}


/*
 *	Returns a statically allocated string containing
 *	the name associated with uid.
 */
char *IDGetNameByUID(uid_t uid)
{
	int i;
	static char rtn_str[NAME_MAX];


	for(i = 0; i < (int)total_id_records; i++)
	{
	    if(id_record[i] == NULL)
		continue;

	    if(uid == id_record[i]->uid)
	    {
		if(id_record[i]->name != NULL)
	            strncpy(rtn_str, id_record[i]->name, 64);
		else
		    return(NULL);
		rtn_str[NAME_MAX - 1] = '\0';

		return(rtn_str);
	    }
	}


        sprintf(rtn_str, "%i", uid);
	return(rtn_str);
}


/*
 *      Returns a statically allocated string containing
 *      the group associated with gid.
 */
char *IDGetGroupByGID(gid_t gid)
{
        int i;
        static char rtn_str[NAME_MAX];


        for(i = 0; i < (int)total_id_records; i++)
        {
            if(id_record[i] == NULL) continue;

            if(gid == id_record[i]->gid)
            {
                if(id_record[i]->name != NULL)
                    strncpy(rtn_str, id_record[i]->name, 64);
                else
                    return(NULL);
                rtn_str[NAME_MAX - 1] = '\0';

                return(rtn_str);
            }
        }


	sprintf(rtn_str, "%i", gid);
        return(rtn_str);
}



/*
 *	Returns a uid associated with name.
 */
uid_t IDGetUIDbyName(char *name)
{
        int i;

        if(name == NULL) return(0);

        /* Is it a number? */
        if(ID_ISSTRINGNUM(name))
            return(atoi(name));


        for(i = 0; i < (int)total_id_records; i++)
        {    
            if(id_record[i] == NULL) continue;
  
            if(!strcmp(id_record[i]->name, name))
                return(id_record[i]->uid);
        }

	return(0);
}


/*
 *      Returns a gid associated with group.
 */
gid_t IDGetGIDbyGroup(char *group)
{
	int i;

	if(group == NULL) return(0);

	/* Is it a number? */
	if(ID_ISSTRINGNUM(group))
	    return(atoi(group));


        for(i = 0; i < (int)total_id_records; i++)
        {
            if(id_record[i] == NULL) continue;

            if(!strcmp(id_record[i]->name, group))
                return(id_record[i]->gid);
        }

        return(0);
}



/* ***************************************************************** */

/*
 *	Allocates a new id record.
 */
int IDAllocate()
{
        int i;
        int new_entry_num;


	/* Sanitize total. */
	if(total_id_records < 0)
	    total_id_records = 0;


	/* Look for NULL entry. */
        for(i = 0; i < total_id_records; i++)
        {
            if(id_record[i] == NULL)
                break;
        }
        if(i < total_id_records)
        {
	    new_entry_num = i;
	}
	else
	{
            new_entry_num = total_id_records;
            total_id_records++;

            id_record = (id_record_struct **)realloc(id_record,
                total_id_records * sizeof(id_record_struct *));
            if(id_record == NULL)
            {
                total_id_records = 0;
                return(-1);
            }
	}

        /* Found one, allocate. */
        id_record[new_entry_num] = (id_record_struct *)calloc(1,
            sizeof(id_record_struct));
        if(id_record[new_entry_num] == NULL)
	{
	    total_id_records--;
            return(-1);
	}


	return(new_entry_num);
}



/*
 *	Procedure to load all uids and gids from the UNIX standard
 *	file containing them.
 */
int IDLoadFromFile(char *filename)
{
	int i;
	struct passwd *pwent;
	struct group *grent;


	/* Note: filename is ignored. */


	/* Free previous listing. */
	IDDeleteAll();


	/* ******************************************************** */
	/* Get UIDs. */

	/* Get first entry. */
	pwent = getpwent();

	while(pwent != NULL)
	{
	    i = IDAllocate();
	    if(i < 0)
		break;

	    id_record[i]->name = StringCopyAlloc(pwent->pw_name);
            id_record[i]->passwd = StringCopyAlloc(pwent->pw_passwd);
	    id_record[i]->uid = pwent->pw_uid;
	    id_record[i]->gid = pwent->pw_gid;
            id_record[i]->cname = StringCopyAlloc(pwent->pw_gecos);
            id_record[i]->dir = StringCopyAlloc(pwent->pw_dir);
            id_record[i]->shell = StringCopyAlloc(pwent->pw_shell);
	    id_record[i]->members = NULL;
/*
printf("UID %i: %s %i %i\n", i,
id_record[i]->name, id_record[i]->uid, id_record[i]->gid);
*/
	    /* Get next password entry entry. */
	    pwent = getpwent();
	}

	/* Close the password entry file. */
	endpwent();


        /* ******************************************************** */
        /* Get GIDs. */

        /* Get first entry. */ 
        grent = getgrent();
        
        while(grent != NULL)
        {
            i = IDAllocate();
            if(i < 0)
                break;

            id_record[i]->name = StringCopyAlloc(grent->gr_name);
            id_record[i]->passwd = StringCopyAlloc(grent->gr_passwd);
            id_record[i]->uid = 0;
            id_record[i]->gid = grent->gr_gid;
            id_record[i]->cname = NULL;
            id_record[i]->dir = NULL;
            id_record[i]->shell = NULL;
	    id_record[i]->members = NULL;	/* We don't need this. */
/*
printf("GID %i: %s %i %i\n", i,
id_record[i]->name, id_record[i]->uid, id_record[i]->gid);
*/
            /* Get next password entry entry. */
            grent = getgrent();
        }

        /* Close the group entry file. */
        endgrent();


	return(0);
}


/*
 *	Deletes all id entries.
 */
void IDDeleteAll()
{
	int i;

	for(i = 0; i < total_id_records; i++)
	{
	    if(id_record[i] == NULL) continue;

	    free(id_record[i]->name);
            free(id_record[i]->passwd);
            free(id_record[i]->cname);
            free(id_record[i]->dir);
            free(id_record[i]->shell);
	    free(id_record[i]->members);

            free(id_record[i]);
	    id_record[i] = NULL;
	}
	free(id_record);
	id_record = NULL;

	total_id_records = 0;


	return;
}
