/*
                       Internal Image Viewers

	Functions:

	int ImgViewerMenuCB(void *client_data, int id_code)
	void ImgViewerRepeatCB()

	int ImgViewerIsAllocated(int n)
	int ImgViewerIsInit(int n)

	int ImgViewerAllocate()
	void ImgViewerReclaimMemory()
	void ImgViewerDelete(int n)
	void ImgViewerDeleteAll()

	int ImgViewerLoadFile(img_viewer_struct *iv, char *filename)
	int ImgViewerInit(img_viewer_struct *iv)
	int ImgViewerResize(img_viewer_struct *iv)
	void ImgViewerBlitView(img_viewer_struct *iv)
	int ImgViewerDraw(img_viewer_struct *iv, int amount)
	int ImgViewerManage(img_viewer_struct *iv, event_t *event)
	int ImgViewerManageAll(event_t *event)
	void ImgViewerMap(img_viewer_struct *iv)
	void ImgViewerUnmap(img_viewer_struct *iv)


	void ImgViewerDestroy(img_viewer_struct *iv)



	---

	These are internal image viewers used for viewing images.

	Imlib is required for these to work.

	Do not confuse image viewers with regular viewers.

*/


#ifdef USE_IMLIB

#include "endeavour.h"


/* Image viewer size values. */
#define ICON_WIDTH	48
#define ICON_HEIGHT	48

#define MIN_WIDTH	240
#define MIN_HEIGHT	200

#define DEF_WIDTH	320
#define DEF_HEIGHT	240

#define CHAR_WIDTH	7
#define CHAR_HEIGHT	14

#define CONSOLE_HEIGHT	50

#define HINT_MESG_RELOAD	"Reload this image (F5)"



/* Menu codes. */
#define IV_MC_NONE		0
#define IV_MC_RELOAD		1

#define IV_MC_ZORIGINAL		10
#define IV_MC_ZNORMAL		11
#define IV_MC_ZFIT		12
#define IV_MC_ZIN		13
#define IV_MC_ZOUT		14


/* Translate and zoom timming constants (in milliseconds). */
#define IV_TRANS_UT	16
#define IV_ZOOM_UT	16
#define IV_TRANS_RATE	4	/* Pixels per unit time. */
#define IV_ZOOM_RATE	0.05	/* Coefficent per unit time. */


/* Translate and zoom repeat records. */
static bool_t	button1_state,
		button2_state,
		button3_state;

static bool_t	shift_key_state,
		alt_key_state,
		ctrl_key_state;

static bool_t	trans_left, trans_right, trans_up, trans_down;
static bool_t	zoom_in, zoom_out, zoom_normal;


/* Milliseconds since last translation or zoom was handled. */
static long	last_handle_time;

/* Last recorded translation position. */
static int	last_trans_x,
		last_trans_y;

/*
 *	Pointer to image viewer that is to handle repeats for or
 *	NULL for none.
 */
static img_viewer_struct	*repeat_iv;




/*
 *	Menu callback.
 */
int ImgViewerMenuCB(void *client_data, int id_code)
{
	int i, len, x, y;
	unsigned int width, height;
	win_attr_t wattr;
	char tmp_path[PATH_MAX + NAME_MAX];


	/* Error checks. */
	if(!IDC() ||
	   (client_data == NULL)
	)
	    return(-1);

	/* ************************************************************** */

	/* Search for valid img_viewer by client_data pointer. */
	for(i = 0; i < (int)total_img_viewers; i++)
	{
	    if(img_viewer[i] == client_data)
		break;
	}
	/* Could not find match? */
	if(i >= (int)total_img_viewers)
	    return(-2);

	/* Check image viewer is allocated and ready for use. */
	if((img_viewer[i]->map_state == 0) ||
           (img_viewer[i]->toplevel == 0) ||
           (img_viewer[i]->view_screen == 0)
	)
	    return(-1);


	/* Handle menu function. */
	switch(id_code)
	{
	  /* ************************************************************ */
	  case IV_MC_RELOAD:
	    if(img_viewer[i]->filename == NULL) break;

	    strncpy(tmp_path, img_viewer[i]->filename, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
	    ImgViewerLoadFile(img_viewer[i], tmp_path);

	    ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
	    break;

          /* ************************************************************ */
	  case IV_MC_ZORIGINAL:
	    if(img_viewer[i]->image_data == NULL) break;
	    if(img_viewer[i]->image_data->data == NULL) break;

            /* Resize window. */
	    width = img_viewer[i]->image_data->width;
	    height = img_viewer[i]->image_data->height;
            OSWResizeWindow(
                img_viewer[i]->toplevel,
		width,
                (int)height + CONSOLE_HEIGHT
	    );
	    img_viewer[i]->zoom = IV_ZOOM_NORMAL;
	    img_viewer[i]->width = 0; /* Fake it into resizing. */
            ImgViewerResize(img_viewer[i]);

	    /* Resize viewscreen image. */
	    if(osw_gui[0].depth == 8)
	        len = (int)width * (int)height * BYTES_PER_PIXEL8;
	    else if(osw_gui[0].depth == 16)
                len = (int)width * (int)height * BYTES_PER_PIXEL16;
            else if((osw_gui[0].depth == 24) ||
                    (osw_gui[0].depth == 32)
            )
                len = (int)width * (int)height * BYTES_PER_PIXEL32;
	    else
		len = (int)width * (int)height * (int)(osw_gui[0].depth >> 3);

            OSWSyncSharedImage(
		img_viewer[i]->view_screen_data,
                img_viewer[i]->view_screen
	    );
	    OSWDestroySharedImage(&img_viewer[i]->view_screen_data);
	    if(
                OSWCreateSharedImage(
		    &img_viewer[i]->view_screen_data,
                    width, height
		)
	    )
		break;

	    memcpy(
		img_viewer[i]->view_screen_data->data,
		img_viewer[i]->image_data->data,
		len
	    );

            ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);

	    break;

          /* ************************************************************ */
          case IV_MC_ZNORMAL:
	    /* Set callback values for zoom normal. */
	    repeat_iv = img_viewer[i];
	    zoom_normal = True;

	    break;

          /* ************************************************************ */
          case IV_MC_ZFIT:
            if(img_viewer[i]->image_data == NULL) break;
            if(img_viewer[i]->image_data->data == NULL) break;
	    if(img_viewer[i]->image_data->width == 0) break;
            if(img_viewer[i]->image_data->height == 0) break;


	    /* Is width of height of view screen smaller? */
	    OSWGetWindowAttributes(img_viewer[i]->view_screen, &wattr);
	    if((int)wattr.width < (int)wattr.height)
	    {
		/* Width is smaller. */
		img_viewer[i]->zoom = (double)wattr.width /
		    (double)img_viewer[i]->image_data->width;
	    }
	    else
            {  
                /* Height is smaller. */
                img_viewer[i]->zoom = (double)wattr.height /
                    (double)img_viewer[i]->image_data->height;
	    }
            img_viewer[i]->view_screen_sb.x_win_pos = 0;
            img_viewer[i]->view_screen_sb.y_win_pos = 0;

            ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
            
            break;
        
          /* ************************************************************ */
          case IV_MC_ZIN:
	    /* Calculate old center position. */
            x = (int)((double)img_viewer[i]->view_screen_sb.x_win_pos +
                (double)(((double)img_viewer[i]->view_screen_data->width /
                img_viewer[i]->zoom) / 2));
            y = (int)((double)img_viewer[i]->view_screen_sb.y_win_pos +
                (double)(((double)img_viewer[i]->view_screen_data->height /
		img_viewer[i]->zoom) / 2));

            img_viewer[i]->zoom += 0.50;
            if(img_viewer[i]->zoom > IV_ZOOM_MAX)
                img_viewer[i]->zoom = IV_ZOOM_MAX;
                        
            if(img_viewer[i]->zoom > IV_ZOOM_NORMAL)
            {
                img_viewer[i]->view_screen_sb.x_win_pos = (int)((double)x -
                    (double)(((double)img_viewer[i]->view_screen_data->width
                    / img_viewer[i]->zoom) / 2));
                img_viewer[i]->view_screen_sb.y_win_pos = (int)((double)y -
                    (double)(((double)img_viewer[i]->view_screen_data->height
                    / img_viewer[i]->zoom) / 2));
            }

            ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);

            break;

          /* ************************************************************ */
          case IV_MC_ZOUT:
            /* Calculate old center position. */
            x = (int)((double)img_viewer[i]->view_screen_sb.x_win_pos +
                (double)(((double)img_viewer[i]->view_screen_data->width /
		img_viewer[i]->zoom) / 2));
            y = (int)((double)img_viewer[i]->view_screen_sb.y_win_pos +
                (double)(((double)img_viewer[i]->view_screen_data->height /
		img_viewer[i]->zoom) / 2));

            img_viewer[i]->zoom -= 0.50;
            if(img_viewer[i]->zoom < IV_ZOOM_MIN)
                img_viewer[i]->zoom = IV_ZOOM_MIN;

            if(img_viewer[i]->zoom > IV_ZOOM_NORMAL)
            {
                img_viewer[i]->view_screen_sb.x_win_pos = (int)((double)x -
                    (double)(((double)img_viewer[i]->view_screen_data->width
                    / img_viewer[i]->zoom) / 2));
                img_viewer[i]->view_screen_sb.y_win_pos = (int)((double)y -
                    (double)(((double)img_viewer[i]->view_screen_data->height
                    / img_viewer[i]->zoom) / 2));
            }
         
            ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);

            break; 
	}



	return(0);
}



/*
 *	Image viewer repeat handler, handles all allocated image viewers.
 */
void ImgViewerRepeatCB()
{
	int i, x, y, x_max, y_max;
	long ct;		/* Current time in milliseconds. */
	double trans_tcomp, zoom_tcomp;		/* Time compensations. */



	/* Get current millitime. */
	ct = MilliTime();
	if(last_handle_time > ct)
	    last_handle_time = ct;

	trans_tcomp = (ct - last_handle_time) / IV_TRANS_UT;
	zoom_tcomp = (ct - last_handle_time) / IV_ZOOM_UT;


	/* Handle all image viewer repeats. */
	for(i = 0; i < (int)total_img_viewers; i++)
	{
	    if(img_viewer[i] == NULL) continue;
	    if(img_viewer[i] != repeat_iv) continue;
	    if(img_viewer[i]->map_state == 0) continue;
	    if(img_viewer[i]->toplevel == 0) continue;


	    /* Translate to the left? */
	    if(trans_left)
	    {
		x = (int)img_viewer[i]->view_screen_sb.x_win_pos;
		x -= (int)(IV_TRANS_RATE * trans_tcomp);
		if(x < 0)
		    x = 0;

		img_viewer[i]->view_screen_sb.x_win_pos = x;

		ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
	    }
	    /* Translate to the right? */
	    if(trans_right)
	    {
		if((img_viewer[i]->image_data != NULL) &&
                   (img_viewer[i]->view_screen_data != NULL)
		)
		    x_max = (int)((int)img_viewer[i]->image_data->width -
                        ((int)img_viewer[i]->view_screen_data->width /
                        img_viewer[i]->zoom));
		else
		    x_max = 0;
		if(x_max < 1)
		    x_max = 1;

                x = img_viewer[i]->view_screen_sb.x_win_pos;
                x += (int)(IV_TRANS_RATE * trans_tcomp);
                if(x > x_max)
		    x = x_max;
        
                img_viewer[i]->view_screen_sb.x_win_pos = x;

                ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
	    }
	    /* Translate up? */
	    if(trans_up)
	    {
                y = img_viewer[i]->view_screen_sb.y_win_pos;
                y -= (int)(IV_TRANS_RATE * trans_tcomp);
                if(y < 0) y = 0;

                img_viewer[i]->view_screen_sb.y_win_pos = y;

                ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
	    }
	    /* Translate down? */
            if(trans_down)
            {
                if((img_viewer[i]->image_data != NULL) &&
                   (img_viewer[i]->view_screen_data != NULL)
                )
                    y_max = (int)((int)img_viewer[i]->image_data->height -
                        ((int)img_viewer[i]->view_screen_data->height /
                        img_viewer[i]->zoom));
                else
                    y_max = 0;
                if(y_max < 1)
		    y_max = 1;

                y = img_viewer[i]->view_screen_sb.y_win_pos;
                y += (int)(IV_TRANS_RATE * trans_tcomp);
                if(y > y_max) y = y_max;

                img_viewer[i]->view_screen_sb.y_win_pos = y;

                ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
            }
	    /* Zoom in? */
            if(zoom_in)
            {
                /* Calculate old center position. */
                x = (int)((double)img_viewer[i]->view_screen_sb.x_win_pos +
                    (double)(((double)img_viewer[i]->view_screen_data->width /
		    img_viewer[i]->zoom) / 2));
                y = (int)((double)img_viewer[i]->view_screen_sb.y_win_pos +
                    (double)(((double)img_viewer[i]->view_screen_data->height /
		    img_viewer[i]->zoom) / 2));

		img_viewer[i]->zoom += (int)(IV_ZOOM_RATE * zoom_tcomp);
		if(img_viewer[i]->zoom > IV_ZOOM_MAX)
		    img_viewer[i]->zoom = IV_ZOOM_MAX;

                if(img_viewer[i]->zoom > IV_ZOOM_NORMAL)
                {
                    img_viewer[i]->view_screen_sb.x_win_pos = (int)((double)x -
                        (double)(((double)img_viewer[i]->view_screen_data->width
                        / img_viewer[i]->zoom) / 2));
                    img_viewer[i]->view_screen_sb.y_win_pos = (int)((double)y -  
                        (double)(((double)img_viewer[i]->view_screen_data->height
                        / img_viewer[i]->zoom) / 2));
                }

                ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
            }
	    /* Zoom out? */
            if(zoom_out)
            {
                /* Calculate old center position. */
                x = (int)((double)img_viewer[i]->view_screen_sb.x_win_pos +
                    (double)(((double)img_viewer[i]->view_screen_data->width /
		    img_viewer[i]->zoom) / 2));
                y = (int)((double)img_viewer[i]->view_screen_sb.y_win_pos +
                    (double)(((double)img_viewer[i]->view_screen_data->height /
		    img_viewer[i]->zoom) / 2));

                img_viewer[i]->zoom -= (int)(IV_ZOOM_RATE * zoom_tcomp);
                if(img_viewer[i]->zoom < IV_ZOOM_MIN)
                    img_viewer[i]->zoom = IV_ZOOM_MIN;

                if(img_viewer[i]->zoom > IV_ZOOM_NORMAL)
                {
                    img_viewer[i]->view_screen_sb.x_win_pos = (int)((double)x -
                        (double)(((double)img_viewer[i]->view_screen_data->width
                        / img_viewer[i]->zoom) / 2));
                    img_viewer[i]->view_screen_sb.y_win_pos = (int)((double)y -
                        (double)(((double)img_viewer[i]->view_screen_data->height
                        / img_viewer[i]->zoom) / 2));
                }

                ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
            }
	    /* Zoom normal? */
            if(zoom_normal)
            {
		if(img_viewer[i]->zoom > IV_ZOOM_NORMAL)
		{
		    img_viewer[i]->zoom -= IV_ZOOM_RATE * zoom_tcomp;
		    if(img_viewer[i]->zoom < IV_ZOOM_NORMAL)
		    {
			img_viewer[i]->zoom = IV_ZOOM_NORMAL;
			zoom_normal = False;	/* Set False here. */
			repeat_iv = NULL;
                        img_viewer[i]->view_screen_sb.x_win_pos = 0;
                        img_viewer[i]->view_screen_sb.y_win_pos = 0;
		    }
		}
		else if(img_viewer[i]->zoom < IV_ZOOM_NORMAL)
                {
                    img_viewer[i]->zoom += IV_ZOOM_RATE * zoom_tcomp;
                    if(img_viewer[i]->zoom > IV_ZOOM_NORMAL)
		    {
                        img_viewer[i]->zoom = IV_ZOOM_NORMAL;
			zoom_normal = False;	/* Set False here. */
			repeat_iv = NULL;
			img_viewer[i]->view_screen_sb.x_win_pos = 0;
                        img_viewer[i]->view_screen_sb.y_win_pos = 0;
		    }
                }

                ImgViewerDraw(img_viewer[i], DRAW_AMOUNT_COMPLETE);
            }
	}

	/* Record last_handle_time. */
	last_handle_time = ct;

	return;
}


/* ************************************************************************** */


int ImgViewerIsAllocated(int n)
{
	if((img_viewer == NULL) ||
           (n < 0) ||
           (n >= (int)total_img_viewers)
	)
	{
	    return(0);
	}
	else if(img_viewer[n] == NULL)
        {
            return(0);
        }
  	else
        {
            return(1);
        }
}



int ImgViewerIsInit(int n)
{
	if(!ImgViewerIsAllocated(n))
	{
	    return(0);
	}
	else if(img_viewer[n]->toplevel == 0)
	{
	    return(0);
	}
	else
	{
            return(1);
        }
}



int ImgViewerAllocate()
{
        int i, n;


        /* Sanitize global total_img_viewers. */
        if((int)total_img_viewers < 0)
            total_img_viewers = 0;
                
             
        /* Check for available entry. */
        for(i = 0; i < (int)total_img_viewers; i++)
        {   
            if(img_viewer[i] == NULL)
                break;
        }
        if(i < (int)total_img_viewers)
        {
	    n = i;
        }
	else
	{
	    n = total_img_viewers;
	    total_img_viewers++;

	    img_viewer = (img_viewer_struct **)realloc(
		img_viewer, total_img_viewers * sizeof(img_viewer_struct *)
	    );
	    if(img_viewer == NULL)
            {
                total_img_viewers = 0;
                return(-1);
            }
	}


        /* Allocate a new image viewer structure. */
        img_viewer[n] = (img_viewer_struct *)calloc(1,
            sizeof(img_viewer_struct));
        if(img_viewer[n] == NULL)
        {
            total_img_viewers--;
            return(-1);
        }


        return(n);
}        



/*
 *	Reclaims image viewer pointers for unallocated image viewers.
 */
void ImgViewerReclaimMemory()
{
        int i, h;


        /* Sanitize global total_img_viewers. */
        if((int)total_img_viewers < 0)
            total_img_viewers = 0;


	/* Get highest number h. */
	for(i = 0, h = -1; i < (int)total_img_viewers; i++)
	{
	    if(ImgViewerIsAllocated(i))
	        h = i;
	}
        if(h < ((int)total_img_viewers - 1))
	{
	    /* Need to reclaim memory. */

            /* Adjust global variable total_img_viewers. */
            total_img_viewers = h + 1;

            /* Reallocate or free pointers. */
            if(total_img_viewers <= 0)
            {
                free(img_viewer);
                img_viewer = NULL;
                total_img_viewers = 0;
            }
            else
            {
                img_viewer = (img_viewer_struct **)realloc(
                    img_viewer,
                    (int)total_img_viewers * sizeof(img_viewer_struct *)
                );
                if(img_viewer == NULL)
                {
                    total_img_viewers = 0;
                }
            }
	}


	return;
}



void ImgViewerDelete(int n)
{
	/* Destroy substructures and resources. */
	if(ImgViewerIsInit(n))
	{
	    ImgViewerDestroy(img_viewer[n]);
	}

	/* Free structure itself. */
	if(ImgViewerIsAllocated(n))
	{
	    free(img_viewer[n]);
	    img_viewer[n] = NULL;
	}


	return;
}



void ImgViewerDeleteAll()
{
	int i;


	for(i = 0; i < (int)total_img_viewers; i++)
	{
	    ImgViewerDelete(i);
	}
	free(img_viewer);
	img_viewer = NULL;

	total_img_viewers = 0;


	return;
}




/* ********************************************************************
 *
 *
 */
int ImgViewerLoadFile(img_viewer_struct *iv, char *filename)
{
        unsigned int width, height;
	char title[PATH_MAX + NAME_MAX + 128];
	char nameonly[PATH_MAX + NAME_MAX];

	pixmap_t pixmap;
        image_t *image;



	/* Error checks. */
	if(!IDC() || (iv == NULL))
	    return(-1);

	if((iv->toplevel == 0) ||
           (iv->view_screen == 0)
	)
	    return(-1);

	if(filename == NULL)
	    return(-1);


        /* ************************************************************** */
	/* Unload previous image and filename. */
	ImgViewerUnload(iv);


	/* ************************************************************** */

	/* Record file name. */
	iv->filename = StringCopyAlloc(filename);

	/* Load image. */
	if(IMLIBOSWLoadFileToImage(filename, &image))
	    return(-1);

        /* Get image values. */
        width = image->width;
        height = image->height;


	/* ************************************************************* */

        /* Set image as actual image data on image viewer. */
        iv->image_data = image;
	image = NULL;

	/* Resize window. */
	OSWResizeWindow(iv->toplevel, width, (int)height + CONSOLE_HEIGHT);
	iv->width = 0;	/* Force resize. */
	OSWGUISync(False);
/*        WidgetCenterWindow(iv->toplevel, WidgetCenterWindowToRoot); */
	ImgViewerResize(iv);


	/* Set new title and icon. */
	sprintf(title, "%s (%ix%i)", filename, width, height);

	strncpy(nameonly, filename, PATH_MAX + NAME_MAX);
	nameonly[PATH_MAX + NAME_MAX - 1] = '\0';
	StripAbsolutePath(nameonly);
        nameonly[PATH_MAX + NAME_MAX - 1] = '\0';

	if(OSWCreateImage(&image, ICON_WIDTH, ICON_HEIGHT))
	    return(-1);
	WidgetResizeImageBuffer(
	    osw_gui[0].depth,
	    (u_int8_t *)image->data,
	    (u_int8_t *)iv->image_data->data,
	    image->width, image->height,
	    iv->image_data->width, iv->image_data->height
	);
	pixmap = OSWCreatePixmapFromImage(image);
	OSWDestroyImage(&image);


	/* WM Properties. */
	OSWSetWindowWMProperties(
	    iv->toplevel,
            title,
            nameonly,
            (pixmap == 0) ? widget_global.std_icon_pm : pixmap,
            False,
            iv->x, iv->y,
            MIN_WIDTH, MIN_HEIGHT,
            osw_gui[0].display_width, osw_gui[0].display_height,
            WindowFrameStyleStandard,
            NULL, 0
	);
        pixmap = 0;     /* Do not destroy pixmap. */


	/* Reset translation and zoom. */
	iv->zoom = 1;
	iv->view_screen_sb.x_win_pos = 0;
        iv->view_screen_sb.y_win_pos = 0;


	return(0);
}



void ImgViewerUnload(img_viewer_struct *iv)
{
        /* Error checks. */
        if(!IDC() || (iv == NULL))
            return;

	/* Free filename. */
	free(iv->filename);
	iv->filename = NULL;


	/* Destroy actual image data. */
	OSWDestroyImage(&iv->image_data);


	/* Do NOT destroy view screen buffer. */


	return;
}



int ImgViewerInit(img_viewer_struct *iv)
{
	int btn_x, btn_y;
        char hotkeystr[PBTN_MAX_HOTKEYS];

        pixmap_t pixmap = 0;
	win_attr_t wattr;


	/* Error checks. */
	if(!IDC() || (iv == NULL))
	    return(-1);


	iv->map_state = 0;
	iv->is_in_focus = 0;
	iv->visibility_state = VisibilityFullyObscured;
	iv->x = 0;
	iv->y = 0;
	iv->width = DEF_WIDTH;
	iv->height = DEF_HEIGHT;
        iv->disabled = False;   

        iv->zoom = 1;

	/* ************************************************************** */

        /* Create toplevel. */
        if(
	    OSWCreateWindow(
	        &iv->toplevel,
	        osw_gui[0].root_win,
	        0, 0,
		iv->width, iv->height
	    )
	)
            return(-1);
        OSWSetWindowInput(iv->toplevel, KeyPressMask |
            KeyReleaseMask | ExposureMask | StructureNotifyMask |
            FocusChangeMask
	);

        WidgetCenterWindow(iv->toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(iv->toplevel, &wattr);
        iv->x = wattr.x;
        iv->y = wattr.y;
        iv->width = wattr.width;
        iv->height = wattr.height;


        /* WM Properties. */
	OSWSetWindowWMProperties(
	    iv->toplevel,
            "No Image Loaded",
            "No Image",
            (pixmap == 0) ? widget_global.std_icon_pm : pixmap,
            True,
            iv->x, iv->y,
            MIN_WIDTH, MIN_HEIGHT,
            osw_gui[0].display_width, osw_gui[0].display_height,
            WindowFrameStyleStandard,
	    NULL, 0
	);
	pixmap = 0;	/* Do not destroy pixmap. */


	/* ********************************************************** */
        /* Viewer window. */
        if(
            OSWCreateWindow( 
                &iv->view_screen,
                iv->toplevel,
                0, 0,
                iv->width, (int)iv->height - CONSOLE_HEIGHT
            )
        )
            return(-1);
	OSWSetWindowInput(iv->view_screen, ButtonPressMask |
	    ButtonReleaseMask | PointerMotionMask | ExposureMask);

	/* Console. */
	if(
            OSWCreateWindow(
                &iv->console,
                iv->toplevel,
                0, (int)iv->height - CONSOLE_HEIGHT,
                iv->width, CONSOLE_HEIGHT
            )
        )
            return(-1);
	OSWSetWindowInput(iv->console, ExposureMask);

	OSWGetWindowAttributes(iv->console, &wattr);
        if(OSWCreatePixmap(&iv->console_buf, wattr.width, wattr.height))
	    return(-1);


	/* Buttons. */
        btn_x = 20;
        btn_y = 10;

        if(
            PBtnInit(
                &iv->close_btn,
                iv->console,
                btn_x, btn_y,
                70, 28,
                "Close",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&iv->close_btn,
		NULL
            )
        )
            return(-1);
        hotkeystr[0] = 0x1b;
        hotkeystr[1] = 'c';
        hotkeystr[2] = '\0';
        PBtnSetHotKeys(&iv->close_btn, hotkeystr);


	btn_x = (int)iv->width - 70 - 20;
        if(     
            PBtnInit(
                &iv->reload_btn,
                iv->console, 
                btn_x, btn_y,
                70, 28,
                "Reload",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&iv->reload_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHintMessage(
            &iv->reload_btn,
            HINT_MESG_RELOAD
        );
        PBtnSetHotKeys(&iv->reload_btn, "r");


	/* Quick menu. */
	if(
	    MenuInit(
		&iv->qmenu,
		osw_gui[0].root_win,
		ImgViewerMenuCB,
		(void *)iv
	    )
	)
	    return(-1);
        MenuAddItem(
            &iv->qmenu,
            "Image Viewer Commands:",
            MENU_ITEM_TYPE_COMMENT,
            NULL,
            IV_MC_NONE,
            -1
        );
        MenuAddItem(
            &iv->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IV_MC_NONE,
            -1
        );
        MenuAddItem(
            &iv->qmenu,
            "Zoom In              Button2",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IV_MC_ZIN,
            -1
        );
        MenuAddItem(
            &iv->qmenu,
            "Zoom Out             Button2",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IV_MC_ZOUT,
            -1
        );
        MenuAddItem(   
            &iv->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IV_MC_NONE,
            -1
        );
        MenuAddItem(
            &iv->qmenu,
            "Zoom 1:1           Backspace",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IV_MC_ZNORMAL,
            -1
        );
        MenuAddItem(
            &iv->qmenu,
            "Zoom To Fit            Alt+F",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IV_MC_ZFIT,
            -1   
        );
        MenuAddItem(
            &iv->qmenu,
            "Original Size          Alt+O",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IV_MC_ZORIGINAL,
            -1
        );
        MenuAddItem(
            &iv->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IV_MC_NONE,
            -1
        );
	MenuAddItem(
	    &iv->qmenu,
	    "Reload                    F5",
	    MENU_ITEM_TYPE_ENTRY,
	    NULL,
	    IV_MC_RELOAD,
	    -1
	);


	return(0);
}



int ImgViewerResize(img_viewer_struct *iv)
{
	int btn_x, btn_y;
        win_attr_t wattr;   
         

        /* Error checks. */
        if(!IDC() ||
           (iv == NULL)
        )
            return(-1);

        if(iv->toplevel == 0)
            return(-1);

        /* ************************************************************** */

        /* Get toplevel attributes. */
        OSWGetWindowAttributes(iv->toplevel, &wattr);

        /* Skip if no change. */
        if(((int)iv->width == wattr.width) &&
           ((int)iv->height == wattr.height)
        )
            return(0);


        /* Set new position and size. */
        iv->x = wattr.x;
        iv->y = wattr.y;
        iv->width = wattr.width;
        iv->height = wattr.height;


	/* Check minimal sizes. */
	if(((int)iv->width < ((2 * 70) + (3 * 20))) ||
           ((int)iv->height < (2 * CONSOLE_HEIGHT))
	)
	{
            OSWSyncSharedImage(
		iv->view_screen_data,
                iv->view_screen
	    );

            /* View screen. */
            OSWMoveResizeWindow(
                iv->view_screen,
                0,
                0,
                iv->width,
                iv->height
            );

            /* Resize image for view_screen_data. */
            OSWDestroySharedImage(&iv->view_screen_data);
            OSWSyncSharedImage(
		iv->view_screen_data,
                iv->view_screen
	    );
            if(
                OSWCreateSharedImage(
                    &iv->view_screen_data,
                    iv->width, iv->height
                )
            )
                return(-1); 

            if(iv->image_data != NULL)
            {
                WidgetResizeImageBuffer(
                    osw_gui[0].depth,
                    iv->view_screen_data->data,		/* Target. */
                    (u_int8_t *)iv->image_data->data,	/* Source. */
                    iv->view_screen_data->width, iv->view_screen_data->height,
                    iv->image_data->width, iv->image_data->height
                );
            }

	    OSWUnmapWindow(iv->console);
	}
	else
	{
            OSWSyncSharedImage(
		iv->view_screen_data,
                iv->view_screen
	    );

            /* View screen. */
            OSWMoveResizeWindow(
                iv->view_screen,
	        0,
	        0,
                iv->width,
                (int)iv->height - CONSOLE_HEIGHT
            );

	    /* Resize image for view_screen_data. */
	    OSWDestroySharedImage(&iv->view_screen_data);
            OSWSyncSharedImage(
		iv->view_screen_data,
                iv->view_screen
	    );
	    if(
	        OSWCreateSharedImage(
		    &iv->view_screen_data,
	            iv->width, (int)iv->height - CONSOLE_HEIGHT
	        )
	    )
	        return(-1);

	    if(iv->image_data != NULL)
	    {
	        WidgetResizeImageBuffer(
		    osw_gui[0].depth,
		    iv->view_screen_data->data,		/* Target. */
		    (u_int8_t *)iv->image_data->data,	/* Source. */
		    iv->view_screen_data->width, iv->view_screen_data->height,
		    iv->image_data->width, iv->image_data->height
	        );
	    }


	    /* Console window. */
	    OSWMapWindow(iv->console);
	    OSWMoveResizeWindow(iv->console,
	        0, (int)iv->height - CONSOLE_HEIGHT,
	        iv->width, CONSOLE_HEIGHT
	    );
	    OSWGetWindowAttributes(iv->console, &wattr);
	    OSWDestroyPixmap(&iv->console_buf);
	    if(OSWCreatePixmap(&iv->console_buf, wattr.width, wattr.height))
	        return(-1);

	    /* Buttons. */
	    btn_x = 20;
	    btn_y = 10;

	    PBtnMap(&iv->close_btn);
	    PBtnMap(&iv->reload_btn);

	    OSWMoveWindow(iv->close_btn.toplevel, btn_x, btn_y);
	    btn_x = (int)iv->width - 70 - 20;
            OSWMoveWindow(iv->reload_btn.toplevel, btn_x, btn_y);
	}


	/* Sync with GUI. */
        OSWGUISync(False);


	return(0);
}



/*
 *	Scales image data to the view screen data according to
 *	translation and zoom.
 */
void ImgViewerBlitView(img_viewer_struct *iv)
{
	int len;
	image_t *src_img;
	shared_image_t *tar_img;

	u_int8_t *src_dp8, *tar_dp8;
	u_int16_t *src_dp16, *tar_dp16;
	u_int32_t *src_dp32, *tar_dp32;

	int src_sk_col, src_sk_row;		/* In * 256. */
	int src_cx, src_cy,			/* In * 256. */
	    src_x, src_y,
            src_w, src_h,
            src_tw, src_th;
	int tar_cx, tar_cy,			/* In * 256. */
	    tar_x, tar_y,
	    tar_w, tar_h,
	    tar_tw, tar_th;
	double zoom;

        int src_width_bytes, tar_width_bytes;	/* NOT * 256. */


	/* Error checks. */
	if(!IDC() ||
           (iv == NULL)
	)
	    return;

	if((iv->view_screen == 0) ||
	   (iv->view_screen_data == NULL) ||
           (iv->image_data == NULL)
	)
	    return;

	if((iv->view_screen_data->data == NULL) ||
           (iv->image_data->data == NULL)
	)
	    return;

	if((iv->view_screen_data->width == 0) ||
           (iv->view_screen_data->height == 0) ||
           (iv->image_data->width == 0) ||
           (iv->image_data->height == 0)
	)
	    return;

	src_img = iv->image_data;
	tar_img = iv->view_screen_data;



	/* *********************************************************** */
	/* Set up values. */

	zoom = iv->zoom;
	if(zoom > IV_ZOOM_MAX)
	    zoom = IV_ZOOM_MAX;
	if(zoom < IV_ZOOM_MIN)
	    zoom = IV_ZOOM_MIN;

	src_sk_col = (int)(256 / zoom);
	src_sk_row = (int)(256 / zoom);

	src_tw = (int)src_img->width * 256;
	src_th = (int)src_img->height * 256;
	tar_tw = (int)tar_img->width * 256;
	tar_th = (int)tar_img->height * 256;

	src_w = (int)(tar_tw / zoom);
        src_h = (int)(tar_th / zoom);
	if(src_w > src_tw) src_w = src_tw;
	if(src_h > src_th) src_h = src_th;

	tar_w = (int)((src_tw * zoom < tar_tw) ? src_tw * zoom : tar_tw);
	tar_h = (int)((src_th * zoom < tar_th) ? src_th * zoom : tar_th);

	tar_x = (tar_w < tar_tw) ? (tar_tw / 2) - (tar_w / 2) : 0;
	if(tar_x < 0) tar_x = 0;
        tar_y = (tar_h < tar_th) ? (tar_th / 2) - (tar_h / 2) : 0;
	if(tar_y < 0) tar_y = 0;

	tar_w += tar_x;
	tar_h += tar_y;

	src_x = iv->view_screen_sb.x_win_pos * 256;
	src_x = ((src_x + src_w) >= src_tw) ? src_tw - src_w : src_x;
	if(src_x < 0) src_x = 0;
        src_y = iv->view_screen_sb.y_win_pos * 256;
        src_y = ((src_y + src_h) >= src_th) ? src_th - src_h : src_y;
        if(src_y < 0) src_y = 0;

	src_w += src_x;
	src_h += src_y;

	src_cx = src_x;
	src_cy = src_y;
	tar_cx = tar_x;
	tar_cy = tar_y;


        /* Clear image if zoom is less than IV_ZOOM_NORMAL. */
        if(((src_tw * iv->zoom) < tar_tw) ||
           ((src_th * iv->zoom) < tar_th)
	)
        {
            len = (int)tar_img->width * (int)tar_img->height *
                (osw_gui[0].depth >> 3);
            memset(tar_img->data, 0x00, len);
        }


	/* 8 bits. */
	if(osw_gui[0].depth == 8)
	{
            src_width_bytes = (int)src_img->width * BYTES_PER_PIXEL8;
            tar_width_bytes = (int)tar_img->width * BYTES_PER_PIXEL8;
                
            while((src_cy < src_h) &&
                  (tar_cy < tar_h)
            )
            {
                src_dp8 = (u_int8_t *)(&src_img->data[
                    ((src_cy >> 8) * src_width_bytes) +
                    ((src_cx >> 8) * BYTES_PER_PIXEL8)
                ]);
                    
                tar_dp8 = (u_int8_t *)(&tar_img->data[
                    ((tar_cy >> 8) * tar_width_bytes) +
                    ((tar_cx >> 8) * BYTES_PER_PIXEL8)
                ]);
                
                *tar_dp8 = *src_dp8;


                /* Increment target x colum. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;
        
                    src_cy += src_sk_row;
                    src_cx = src_x;
                }
         
                /* Increment source x colum. */
                src_cx += src_sk_col;
                /* Go to next source line? */
                if(src_cx >= src_w)  
                {
                    tar_cy += 256;
                    tar_cx = tar_x;
         
                    src_cy += src_sk_row;
                    src_cx = src_x;
                }  
            }
	}
	/* 16 bits. */
	else if(osw_gui[0].depth == 16)
	{
	    src_width_bytes = (int)src_img->width * BYTES_PER_PIXEL16;
	    tar_width_bytes = (int)tar_img->width * BYTES_PER_PIXEL16;

	    while((src_cy < src_h) &&
                  (tar_cy < tar_h)
	    )
	    {
		src_dp16 = (u_int16_t *)(&src_img->data[
		    ((src_cy >> 8) * src_width_bytes) +
		    ((src_cx >> 8) * BYTES_PER_PIXEL16)
		]);

		tar_dp16 = (u_int16_t *)(&tar_img->data[
                    ((tar_cy >> 8) * tar_width_bytes) +
                    ((tar_cx >> 8) * BYTES_PER_PIXEL16)
                ]);

		*tar_dp16 = *src_dp16;


                /* Increment target x colum. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

		    src_cy += src_sk_row;
		    src_cx = src_x;
                }

		/* Increment source x colum. */
		src_cx += src_sk_col;
                /* Go to next source line? */
		if(src_cx >= src_w)
		{
                    tar_cy += 256;
                    tar_cx = tar_x;

		    src_cy += src_sk_row;
		    src_cx = src_x;
		}
	    }
	}
	/* 24 or 32 bits. */
	else if((osw_gui[0].depth == 24) ||
                (osw_gui[0].depth == 32)
	)
	{
            src_width_bytes = (int)src_img->width * BYTES_PER_PIXEL32;
            tar_width_bytes = (int)tar_img->width * BYTES_PER_PIXEL32;
                
            while((src_cy < src_h) &&
                  (tar_cy < tar_h)
            )
            {    
                src_dp32 = (u_int32_t *)(&src_img->data[
                    ((src_cy >> 8) * src_width_bytes) +
                    ((src_cx >> 8) * BYTES_PER_PIXEL32)
                ]); 
         
                tar_dp32 = (u_int32_t *)(&tar_img->data[
                    ((tar_cy >> 8) * tar_width_bytes) +
                    ((tar_cx >> 8) * BYTES_PER_PIXEL32)
                ]);
         
                *tar_dp32 = *src_dp32;
         
         
                /* Increment target x colum. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

                    src_cy += src_sk_row;
                    src_cx = src_x;
                }
        
                /* Increment source x colum. */
                src_cx += src_sk_col;
                /* Go to next source line? */
                if(src_cx >= src_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;
         
                    src_cy += src_sk_row;
                    src_cx = src_x;
                }
            }
	}


	return;
}



int ImgViewerDraw(img_viewer_struct *iv, int amount)
{
	int x, len;
        win_attr_t wattr;
	char stringa[256];


	if(iv == NULL)
	    return(-1);


	/* Map as needed. */
	if(!iv->map_state)
	{
            OSWMapRaised(iv->toplevel);   
            OSWMapWindow(iv->view_screen);

            if(((int)iv->width < ((2 * 70) + (3 * 20))) ||
               ((int)iv->height < (2 * CONSOLE_HEIGHT))
	    )
	    {
		OSWUnmapWindow(iv->console);
	    }
	    else
	    {
	        OSWMapWindow(iv->console);

	        PBtnMap(&iv->close_btn);
	        PBtnMap(&iv->reload_btn);
	    }
	    MenuUnmap(&iv->qmenu);

	    iv->visibility_state = VisibilityUnobscured;
	    iv->map_state = 1;
	}


	/* View screen. */
	if(1)
	{
	    ImgViewerBlitView(iv);
	    OSWPutSharedImageToDrawable(
		iv->view_screen_data,
		(drawable_t)iv->view_screen
	    );
	}

	/* Console. */
	if(1)
	{
	    OSWGetWindowAttributes(iv->console, &wattr);
	    WidgetPutImageTile(
		(drawable_t)iv->console_buf,
		widget_global.std_bkg_img,
		wattr.width, wattr.height
	    );

	    /* Draw geometry or op label as needed. */
	    if(((int)iv->width > (60 * CHAR_WIDTH)) &&
               (iv->view_screen_data != NULL)
	    )
	    {
	        if(button1_state || button2_state)
		{
		    sprintf(stringa, "%ix%i+%i+%i (%.0f%%)",
                        (int)((double)iv->view_screen_data->width / iv->zoom), 
                        (int)((double)iv->view_screen_data->height / iv->zoom),
		        (int)iv->view_screen_sb.x_win_pos,
		        (int)iv->view_screen_sb.y_win_pos,
		        (double)iv->zoom * 100
		    );
		    len = strlen(stringa);
		    x = ((int)wattr.width / 2) - ((len * CHAR_WIDTH) / 2);
		}
		else
		{
		    sprintf(stringa, "Button1 = Move  Button2 = Zoom");
		    len = strlen(stringa);
                    x = ((int)wattr.width / 2) - ((len * CHAR_WIDTH) / 2);
		}

		OSWSetFgPix(widget_global.normal_text_pix);
		OSWDrawString(
		    iv->console_buf,
		    x, ((int)wattr.height / 2) + 5,
		    stringa
		);
	    }

	    WidgetFrameButtonPixmap(
                iv->console_buf,
		False,
		wattr.width, wattr.height,
		widget_global.surface_highlight_pix,
		widget_global.surface_shadow_pix
	    );

	    OSWPutBufferToWindow(iv->console, iv->console_buf);
	}


	return(0);
}



int ImgViewerManage(img_viewer_struct *iv, event_t *event)
{
	double zoom;
	int i, x, y, root_x, root_y;
	int events_handled = 0;
	char tmp_path[PATH_MAX + NAME_MAX];
	char *strptr;
	endeavour_win_struct *edw;


	if(!IDC() || (iv == NULL) || (event == NULL))
	    return(0);

	switch(event->type)
	{
          /* ********************************************************* */
          case ButtonPress:
	    if(event->xany.window == iv->view_screen)
            {
		if(event->xbutton.button == Button1) button1_state = True;
		if(event->xbutton.button == Button2) button2_state = True;
		if(event->xbutton.button == Button3) button3_state = True;

		last_trans_x = event->xbutton.x;
		last_trans_y = event->xbutton.y;


                if(button1_state && !alt_key_state &&
                   (endeavour_cursor.translate != NULL)
		)
		    OSWSetWindowCursor(iv->view_screen,
			endeavour_cursor.translate->cursor);
		if((button2_state || (button1_state && alt_key_state)) &&
                   (endeavour_cursor.zoom != NULL)
		)
		    OSWSetWindowCursor(iv->view_screen,
			endeavour_cursor.zoom->cursor);

		/* Map the quick menu. */
		if(event->xbutton.button == Button3)
		{
		    OSWGetPointerCoords(iv->toplevel,
			&root_x, &root_y,
			&x, &y
		    );
		    OSWMoveWindow(iv->qmenu.toplevel, root_x, root_y);
		    iv->qmenu.x = root_x;
		    iv->qmenu.y = root_y;

		    MenuMap(&iv->qmenu);

                    events_handled++;
		    return(events_handled);
		}
	    }
	    break;

          /* ********************************************************* */
	  case ButtonRelease:
	    if(event->xany.window == iv->view_screen)
	    {
		/* Local button states. */
                if(event->xbutton.button == Button1) button1_state = False;
                if(event->xbutton.button == Button2) button2_state = False;
                if(event->xbutton.button == Button3) button3_state = False;

                last_trans_x = event->xbutton.x;
                last_trans_y = event->xbutton.y;

		if(!button1_state && !button2_state)
		    OSWSetWindowCursor(iv->view_screen, osw_gui[0].std_cursor);

		/* Global drag record check. */
		if(drag_record[0].state)
		{
		    if(drag_record[0].origin_ptr_type == POINTER_TYPE_EDW)
		        edw = (endeavour_win_struct *)drag_record[0].origin_ptr;
		    else
			edw = NULL;

		    if(edw != NULL)
		    {
			if((edw->total_sel_files > 0) &&
			   (edw->loc_prompt.buf != NULL)
			)
			{
			    i = edw->sel_file[0];
			    if((i >= 0) && (i < (int)edw->total_files))
			    {
				strptr = NULL;
				if(edw->file[i] != NULL)
				{
				    strptr = PrefixPaths(
					edw->loc_prompt.buf,
					edw->file[i]->name
				    );
				}
				strncpy(tmp_path,
				    (strptr == NULL) ? "/" : strptr,
				    PATH_MAX + NAME_MAX
				);
				tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

				ImgViewerLoadFile(iv, tmp_path);
				ImgViewerMap(iv);
			    }
			}
		    }
		}

                events_handled++;
	    }

            break;

          /* ********************************************************* */
          case KeyPress:
	    if(iv->is_in_focus == 0)
          	return(events_handled);

	    /* Modifier keys. */
	    if((event->xkey.keycode == osw_keycode.shift_left) ||
               (event->xkey.keycode == osw_keycode.shift_right)
	    )
		shift_key_state = True;
            else if((event->xkey.keycode == osw_keycode.alt_left) ||
                    (event->xkey.keycode == osw_keycode.alt_right)
            )
                alt_key_state = True;
            else if((event->xkey.keycode == osw_keycode.ctrl_left) ||
                    (event->xkey.keycode == osw_keycode.ctrl_right)
            )
                ctrl_key_state = True;


	    /* Change cursor for alt_key_state change. */
	    if(alt_key_state && button1_state &&
               (endeavour_cursor.zoom != NULL)
	    )
		OSWSetWindowCursor(iv->view_screen,
                        endeavour_cursor.zoom->cursor);
	    else if(!alt_key_state && button1_state &&
                    (endeavour_cursor.translate != NULL)
	    )
                OSWSetWindowCursor(iv->view_screen,
                    endeavour_cursor.translate->cursor);


	    /* Reload F5. */
	    if(event->xkey.keycode == osw_keycode.f5)
	    {
		iv->reload_btn.state = PBTN_ARMED;
		PBtnDraw(&iv->reload_btn);

                OSWKBAutoRepeatOff();

		events_handled++;
		return(events_handled);
	    }
	    /* Zoom to fit. */
            else if((event->xkey.keycode == osw_keycode.alpha_f) &&
	            alt_key_state
	    )
            {
	        ImgViewerMenuCB(iv, IV_MC_ZFIT);

		events_handled++;
                return(events_handled);
	    }
            /* Original size. */
            else if((event->xkey.keycode == osw_keycode.alpha_o) &&
                    alt_key_state
            )
            {
                ImgViewerMenuCB(iv, IV_MC_ZORIGINAL);

                events_handled++;
                return(events_handled);
            }

	    /* Up. */
	    else if(event->xkey.keycode == osw_keycode.cursor_up)
	    {
		trans_up = True;
		repeat_iv = iv;
		OSWKBAutoRepeatOff();

		events_handled++;
		return(events_handled);
	    }
            /* Down. */
            else if(event->xkey.keycode == osw_keycode.cursor_down)
            {
                trans_down = True;
                repeat_iv = iv;
                OSWKBAutoRepeatOff();

                events_handled++;
		return(events_handled);
            }
            /* Left. */
            else if(event->xkey.keycode == osw_keycode.cursor_left)
            {
                trans_left = True;
                repeat_iv = iv;
                OSWKBAutoRepeatOff();

                events_handled++;
		return(events_handled);
            }
            /* Right. */
            else if(event->xkey.keycode == osw_keycode.cursor_right)
            {
                trans_right = True;
                repeat_iv = iv;
                OSWKBAutoRepeatOff();
                
                events_handled++;
		return(events_handled);
            }
            /* Zoom out. */
            else if((event->xkey.keycode == osw_keycode.minus) ||
                    (event->xkey.keycode == osw_keycode.np_minus)
	    )
            {
		zoom_out = True;
                repeat_iv = iv;
                OSWKBAutoRepeatOff();

                events_handled++;
		return(events_handled);
            }
            /* Zoom in. */
            else if((event->xkey.keycode == osw_keycode.equal) ||
                    (event->xkey.keycode == osw_keycode.np_add)
	    )
            {
                zoom_in = True;
                repeat_iv = iv;
                OSWKBAutoRepeatOff();

                events_handled++;
		return(events_handled);
            }
            /* Zoom normal. */
            else if((event->xkey.keycode == osw_keycode.backspace) ||
                    (event->xkey.keycode == osw_keycode.np_5)
	    )
            {
                zoom_normal = True;
                repeat_iv = iv;
                OSWKBAutoRepeatOff();


                events_handled++;
                return(events_handled);
            }
	    break;

          /* ********************************************************* */
          case KeyRelease:
            if(iv->is_in_focus == 0)
                return(events_handled);

            /* Modifier keys. */
            if((event->xkey.keycode == osw_keycode.shift_left) ||
               (event->xkey.keycode == osw_keycode.shift_right)
            )
                shift_key_state = False;
            else if((event->xkey.keycode == osw_keycode.alt_left) ||
                    (event->xkey.keycode == osw_keycode.alt_right)
            )
                alt_key_state = False;
            else if((event->xkey.keycode == osw_keycode.ctrl_left) ||
                    (event->xkey.keycode == osw_keycode.ctrl_right)
            )
                ctrl_key_state = False;


            /* Change cursor for alt_key_state change. */
            if(alt_key_state && button1_state &&
               (endeavour_cursor.zoom != NULL)
            )
                OSWSetWindowCursor(iv->view_screen,
                        endeavour_cursor.zoom->cursor);
            else if(!alt_key_state && button1_state &&
                    (endeavour_cursor.translate != NULL)
            )
                OSWSetWindowCursor(iv->view_screen,
                    endeavour_cursor.translate->cursor);
 

	    /* F5. */
            if(event->xkey.keycode == osw_keycode.f5)
            {
                iv->reload_btn.state = PBTN_UNARMED;
                PBtnDraw(&iv->reload_btn);

                if(iv->filename != NULL)
                {
                    strncpy(tmp_path, iv->filename, PATH_MAX + NAME_MAX);
                    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
          
                    ImgViewerLoadFile(iv, tmp_path);
                }


		OSWKBAutoRepeatOn();

                events_handled++;
                return(events_handled);
            }

            /* Up. */
            else if(event->xkey.keycode == osw_keycode.cursor_up)
            {
                trans_up = False;
                events_handled++;
            }
            /* Down. */
            else if(event->xkey.keycode == osw_keycode.cursor_down)
            {
                trans_down = False;
                events_handled++;
            }
            /* Left. */
            else if(event->xkey.keycode == osw_keycode.cursor_left)
            {
                trans_left = False;
                events_handled++;
            }
            /* Right. */
            else if(event->xkey.keycode == osw_keycode.cursor_right)
            {
                trans_right = False;
                events_handled++;
            }
            /* Zoom out. */
            else if((event->xkey.keycode == osw_keycode.minus) ||
                    (event->xkey.keycode == osw_keycode.np_minus)
            )
            {
                zoom_out = False;
                events_handled++;
            }
            /* Zoom in. */
            else if((event->xkey.keycode == osw_keycode.equal) ||
                    (event->xkey.keycode == osw_keycode.np_add)
            )
            {
                zoom_in = False;
                events_handled++;
            }
            /* Zoom normal. */
            else if((event->xkey.keycode == osw_keycode.backspace) ||
                    (event->xkey.keycode == osw_keycode.np_5)
            )
            {
/*		Do not set this to False, it will be set by repeat callback
                function.

                zoom_normal = False;
*/
                OSWKBAutoRepeatOn();

                events_handled++;
            }

	    /*   Turn keyboard auto repeat back on and repeat_iv to NULL
             *   when all states are False.
	     */
            if(!trans_up &&
               !trans_down &&
               !trans_left &&
               !trans_right &&
               !zoom_in &&
               !zoom_out
	       /* Do not check zoom_normal. */
            )
            {
                OSWKBAutoRepeatOn();
            }
            if(!trans_up &&
               !trans_down &&
               !trans_left && 
               !trans_right &&
               !zoom_in &&
               !zoom_out &&
	       !zoom_normal	/* Yes check zoom_normal here. */
            )
	    {
		repeat_iv = NULL;
	    }

            break;

          /* ********************************************************* */
          case MotionNotify:
	    if(event->xany.window == iv->view_screen)
	    {
                /* Translate. */
                if(button1_state &&
                   !alt_key_state &&
		   (iv->image_data != NULL) &&
                   (iv->view_screen_data != NULL)
		)
		{
		    x = (int)iv->image_data->width -
			(int)((int)iv->view_screen_data->width / iv->zoom);
                    y = (int)iv->image_data->height -
                        (int)((int)iv->view_screen_data->height / iv->zoom);

		    iv->view_screen_sb.x_win_pos += (int)(
			(double)(last_trans_x - event->xmotion.x) / iv->zoom
		    );
		    iv->view_screen_sb.y_win_pos += (int)(
			(double)(last_trans_y - event->xmotion.y) / iv->zoom
		    );

		    if(iv->view_screen_sb.x_win_pos >= x)
			iv->view_screen_sb.x_win_pos = x - 1;
		    if(iv->view_screen_sb.x_win_pos < 0)
                        iv->view_screen_sb.x_win_pos = 0;

                    if(iv->view_screen_sb.y_win_pos >= y)
                        iv->view_screen_sb.y_win_pos = y - 1;
                    if(iv->view_screen_sb.y_win_pos < 0)  
                        iv->view_screen_sb.y_win_pos = 0;

		    /* Update last translation positions. */
		    last_trans_x = event->xmotion.x;
		    last_trans_y = event->xmotion.y;

		    OSWPurgeTypedEvent(MotionNotify);

		    events_handled++;
		}
                /* Zoom. */
                if((button2_state || (button1_state && alt_key_state)) &&
                   (iv->image_data != NULL) &&
                   (iv->view_screen_data != NULL)
                )
	        {
		    if(iv->view_screen_data->height != 0)
		    {
			/* Calculate old center ratios. */
			x = (int)((double)iv->view_screen_sb.x_win_pos +
			    (double)(((double)iv->view_screen_data->width /
			    iv->zoom) / 2));
                        y = (int)((double)iv->view_screen_sb.y_win_pos +
                            (double)(((double)iv->view_screen_data->height /
			    iv->zoom) / 2));

			/* Calculate delta zoom (can be negative). */
			zoom = (last_trans_y - event->xmotion.y) *
                         (double)(1 / (double)iv->view_screen_data->height)
                         * 3;

		        iv->zoom -= zoom;

			/*   Adjust translation position to reflect zoomed
                         *   transferance.
			 */
			iv->view_screen_sb.x_win_pos = (int)((double)x -
                            (double)(((double)iv->view_screen_data->width
			    / iv->zoom) / 2));
                        iv->view_screen_sb.y_win_pos = (int)((double)y -
                            (double)(((double)iv->view_screen_data->height
			    / iv->zoom) / 2));

			if(iv->view_screen_sb.x_win_pos < 0)
			    iv->view_screen_sb.x_win_pos = 0;
			if(iv->view_screen_sb.y_win_pos < 0)
			    iv->view_screen_sb.y_win_pos = 0;
		    }
		    else
		    {
			iv->zoom = IV_ZOOM_NORMAL;
		    }
		    if(iv->zoom < IV_ZOOM_MIN) iv->zoom = IV_ZOOM_MIN;
		    if(iv->zoom > IV_ZOOM_MAX) iv->zoom = IV_ZOOM_MAX;


                    /* Update last translation positions. */
                    last_trans_x = event->xmotion.x; 
                    last_trans_y = event->xmotion.y;

                    OSWPurgeTypedEvent(MotionNotify);
            
                    events_handled++;
                }


	    }
	    break;

          /* ********************************************************* */
	  case Expose:
	    if((event->xany.window == iv->toplevel) ||
               (event->xany.window == iv->view_screen) ||
               (event->xany.window == iv->console)
	    )
	    {
		events_handled++;
	    }
	    break;

          /* ********************************************************* */
          case ConfigureNotify:
            if(event->xany.window == iv->toplevel)
            {
                ImgViewerResize(iv);

                events_handled++;
            }
            break;

          /* ********************************************************* */
          case FocusIn:
            if(event->xany.window == iv->toplevel)
            {
                iv->is_in_focus = 1;
                events_handled++;
            }
            else
            {
                iv->is_in_focus = 0;
                /* Do not report any event for focus out. */
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case FocusOut:
            if(event->xany.window == iv->toplevel)
            {
                iv->is_in_focus = 0;
                events_handled++;
            }   
            return(events_handled);
            break;
                
          /* ********************************************************* */
          case VisibilityNotify:
            if(event->xany.window == iv->toplevel)
            {
                iv->visibility_state = event->xvisibility.state;
          
                events_handled++;
                return(events_handled);
            }   
            break;

          /* ********************************************************* */
          case ClientMessage:
            if(OSWIsEventDestroyWindow(iv->toplevel, event))
            {
                ImgViewerUnmap(iv);

                events_handled++;
                return(events_handled);
            }
            break;
        
          /* ********************************************************* */
          case LeaveNotify:
            break;
	}

	/* Redraw as needed. */
	if(events_handled > 0)
	{
	    ImgViewerDraw(iv, DRAW_AMOUNT_COMPLETE);
	}


	/* ************************************************************ */
	if(events_handled == 0)
	{
	    events_handled += PBtnManage(&iv->close_btn, event);

            if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
               (event->type == KeyRelease))
            )
            {
		ImgViewerUnmap(iv);
                return(events_handled);
            }
	}

        if(events_handled == 0)
        {
            events_handled += PBtnManage(&iv->reload_btn, event);

	    if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
               (event->type == KeyRelease))
	    )
	    {
		if(iv->filename != NULL)
		{
		    strncpy(tmp_path, iv->filename, PATH_MAX + NAME_MAX);
		    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

		    ImgViewerLoadFile(iv, tmp_path);
		}
	    }
        }

	/* Quick menu. */
        if(events_handled == 0)
        {
	    x = iv->qmenu.map_state;

            events_handled += MenuManage(&iv->qmenu, event);

	    /* Unmap as needed. */
	    if((events_handled <= 0) &&
               (event->type == ButtonRelease)
	    )
	    {
		MenuUnmap(&iv->qmenu);
	    }

	    /* Check if it just unmapped. */
            if((x == 1) &&
               (iv->qmenu.map_state == 0)
	    )
	    {
		OSWPurgeAllEvents();
                ImgViewerDraw(iv, DRAW_AMOUNT_COMPLETE);
	    }
        }


	return(events_handled); 
}



int ImgViewerManageAll(event_t *event)
{
        static int i, events_handled;
  

        /* Error checks. */
        if(!IDC() || 
           (event == NULL)
        )
            return(0);


        /* ********************************************************** */
        events_handled = 0;

        for(i = 0; i < (int)total_img_viewers; i++)
        {
            /* Skip unallocated ones. */
            if(img_viewer[i] == NULL)
                continue;

            /* ClientMessage must be handled here. */
            if((event->type == ClientMessage) &&
               OSWIsEventDestroyWindow(img_viewer[i]->toplevel, event)
            )
            {
                /* Delete viewer and reclaim memory. */
                ImgViewerDelete(i);
                ImgViewerReclaimMemory();     

                OSWPurgeAllEvents();
                events_handled++;
                break;
            }

            /* Send event to viewer widget management for handling. */
            events_handled += ImgViewerManage(img_viewer[i], event);


            /*   If viewer unmapped while it was being managed, we should
             *   just delete it.
             */
            if(img_viewer[i]->map_state == 0)
            {
                ImgViewerDelete(i);
                ImgViewerReclaimMemory();     /* Reclaim mem after each delete. */
            }
        }


        return(events_handled);
}



void ImgViewerMap(img_viewer_struct *iv)
{
	if(!IDC() || (iv == NULL))
	    return;

        UnfocusAllTopLvlWins();

	iv->map_state = 0;
	ImgViewerDraw(iv, DRAW_AMOUNT_COMPLETE);

	iv->is_in_focus = 1;


	return;
}



void ImgViewerUnmap(img_viewer_struct *iv)
{
        if(!IDC() || (iv == NULL))
            return;


	OSWUnmapWindow(iv->toplevel);
	iv->map_state = 0;
	iv->visibility_state = VisibilityFullyObscured;
	iv->is_in_focus = 0;

	return;
}


void ImgViewerDestroy(img_viewer_struct *iv)
{
	if(iv == NULL)
	    return;

	/* Unset repeat_iv if the one being destroyed is it. */
	if(iv == repeat_iv)
	    repeat_iv = NULL;


	iv->map_state = 0;
        iv->visibility_state = VisibilityFullyObscured;
        iv->x = 0;
	iv->y = 0;
        iv->is_in_focus = 0;
        iv->width = 0;
	iv->height = 0;
        iv->disabled = False;

	free(iv->filename);
	iv->filename = NULL;


	if(IDC())
	{
	    MenuDestroy(&iv->qmenu);

	    PBtnDestroy(&iv->close_btn);
	    PBtnDestroy(&iv->reload_btn);

	    OSWDestroyWindow(&iv->console);
	    OSWDestroyPixmap(&iv->console_buf);

	    OSWSyncSharedImage(
		iv->view_screen_data,
		iv->view_screen
	    );
	    OSWDestroyWindow(&iv->view_screen);
	    OSWDestroySharedImage(&iv->view_screen_data);

	    OSWDestroyWindow(&iv->toplevel);
            OSWDestroyPixmap(&iv->toplevel_buf);

	    OSWDestroyImage(&iv->image_data);
	}


	return;
}


#endif /* USE_IMLIB */
