/*
                           Recycle Bin IO

	Macros::

	void RB_DELETE_REALOBJECT(char *path)
	char *GETGLOBALRECYCLEBIN(char *username)
	int SETUPRECYCLEBIN(const char *recbin)
	int GETENTRYNUMBYNAME(char *path, char *recbin)


	Recycle Bin IO:

	int RecycleAddObject(const char *path, const char *recbin)
	int RecycleRemoveObject(int entry_num, char *recbin)
	int RecycleRecoverObject(int entry_num, char *recbin, char *path)


	Front ends:

	void RecycleDoPurgeObject(char *string, char *recbin)
	int RecycleDoRecycleObject(const char *path, const char *recbin)
	int RecycleDoRestoreObject(char *string, char *recbin, char *path)  
	object_struct **RecycleGetObjectListing(char *recbin, int *nobjs)


	---

	Note: The term `real object' herein reffers to an non-recycled
	object NOT in the recycle bin.

 */

#include "endeavour.h"



/*
 *	Removes real object specified in path.  Directories will be left
 *	alone (not deleted).   If the object is a regular file, it will
 *	be purged first then unlink()'ed.
 *
 *	WARNING: This function does NOT add the object into the
 *	recycle bin.
 */
void RB_DELETE_REALOBJECT(char *path)
{
	register int i;
	register off_t filesize;
	register FILE *fp;
	char tmp_path[PATH_MAX + NAME_MAX];
	struct stat stat_buf;


	/* Error checks. */
	if(path == NULL) return;

        strncpy(tmp_path, path, PATH_MAX + NAME_MAX);
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
        StringStripSpaces(tmp_path);


	/* Get stats and check if it exists. */
	if(lstat(tmp_path, &stat_buf)) return;

	/* Get file size. */
	filesize = stat_buf.st_size;


	/* ****************************************************** */

	/* Link. */
	if(S_ISLNK(stat_buf.st_mode))
	{
	    unlink(tmp_path);
	}
	/* Regular file. */
	else if(S_ISREG(stat_buf.st_mode))
	{
	    /* Open file in read/write for purging. */
	    fp = fopen(tmp_path, "r+");
	    if(fp != NULL)
	    {
	        /* Stream is at beginning of file. */

	        /* `purge' file. */
	        for(i = 0; i < filesize; i++)
	        {
		    fputc(0x00, fp);
	        }

	        fclose(fp);
	    }

	    /* Delete the file. */
	    unlink(tmp_path);
	}
	/* Directories, leave alone. */
	else if(S_ISDIR(stat_buf.st_mode))
	{
	    /* Do nothing. */
	}
	/* FIFO pipe. */
	else if(S_ISFIFO(stat_buf.st_mode))
	{
	    unlink(tmp_path);
	}
	/* Socket. */
	else if(S_ISSOCK(stat_buf.st_mode))
        {
            unlink(tmp_path);
        }


	return;
}



/*
 *	Get the full path to the global recycle bin.
 *	Return string is statically allocated and can be NULL on error.
 */
char *GETGLOBALRECYCLEBIN(char *username)
{
	static char tmp_path[PATH_MAX + NAME_MAX];


	/* Get username as needed from enviroment setting. */
	if(username == NULL)
	{
            username = getenv("USERNAME");
            if(username == NULL)
		username = getenv("USER");
	    /* Give up if cannot get user's name from enviroment. */
	    if(username == NULL)
		return(NULL);
	}

	/* Prefix global path with username. */
	strncpy(
	    tmp_path,
	    PrefixPaths(RECYCLE_GLOBAL_DIR, username),
	    PATH_MAX + NAME_MAX
	);

	/* Tail recycle bin name to path. */
	strncat(
	    tmp_path,
	    RECYCLE_BIN_FILE_NAME,
	    PATH_MAX + NAME_MAX - strlen(tmp_path) - 1
	);
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


	return(tmp_path);
}



/*
 *	This is just some init stuff that needs to be done
 *	each time a Recycle* function is called, this is used interanlly
 *	and there is no need to uninit or shutdown this.
 */
int SETUPRECYCLEBIN(const char *recbin)
{
	int status;
	char parent[PATH_MAX];
	char *strptr;
	mode_t mode;


	if(recbin == NULL)
	    return(-1);


        /* Get parent path. */
	strptr = GetParentDir(recbin);
	strncpy(
	    parent,
	    (strptr == NULL) ? RECYCLE_GLOBAL_DIR : strptr,
	    PATH_MAX
	);
	parent[PATH_MAX - 1] = '\0';

	/* If parent exists, it must be a dir. */
	if(access(parent, F_OK) == 0)
	{
	    if(!ISPATHDIR(parent))
	    {
		fprintf(stderr,
		    "%s: Not a directory.\n",
		    parent
		);
		fprintf(stderr,
"\nThe parent directory for the recycle bin does not appear to\n\
be a directory.   Check to make sure that your directory structures\n\
and all refferances to them (eg: configuration files) are defined\n\
correctly.\n"
		);
	        return(-1);
	    }
	}
	else
	{
	    /* Recycle bin dir does not exist, create it. */

	    /* Set up permissions for owner rwx only. */
	    mode = 0;
	    mode |= S_IRUSR;
            mode |= S_IWUSR;
            mode |= S_IXUSR;
	    status = mkdir(parent, mode);

	    return(status);
	}




	return(0);
}



int GETENTRYNUMBYNAME(char *path, char *recbin)
{
	FILE *fp;
	struct stat stat_buf;
	off_t filesize;
	u_int64_t chunksize;
	long next_pos;
	long cur_pos;
	char tmp_path[PATH_MAX + NAME_MAX];

	int entry_num;


	if(path == NULL)
	    return(-1);

	fp = fopen(recbin, "rb");
	if(fp == NULL)
	    return(-1);

	fstat(fileno(fp), &stat_buf);
	filesize = stat_buf.st_size;

	next_pos = 0;
	cur_pos = 0;
	entry_num = 0;
	while(next_pos < filesize)
	{
	    /* Read header/attributes chunk. */
	    fseek(fp, cur_pos, SEEK_SET);
	    fread(&chunksize, sizeof(u_int64_t), 1, fp);
	    next_pos += chunksize;

	    /* Get name. */
	    fseek(fp, cur_pos + 88, SEEK_SET);
	    fread(tmp_path, sizeof(char), PATH_MAX + NAME_MAX, fp);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
	    if(!strcmp(path, tmp_path))
	    {
		fclose(fp); fp = NULL;
		return(entry_num);
	    }
	    cur_pos = next_pos;


	    /* Read contencts chunk. */
	    fseek(fp, cur_pos, SEEK_SET);
            fread(&chunksize, sizeof(u_int64_t), 1, fp);
            next_pos += chunksize;

	    cur_pos = next_pos;

	    /* Increment entry number. */
	    entry_num++;
	}

	fclose(fp); fp = NULL;


	return(-1);
}



/* ********************************************************************** */

/*
 *	Adds real object path into the recycle bin recbin
 *	and then erases it from disk if possible.
 */
int RecycleAddObject(const char *path, const char *recbin)
{
	FILE *src_fp;
	FILE *tar_fp;

	char src_path[PATH_MAX + NAME_MAX];
	char tar_path[PATH_MAX + NAME_MAX];

	struct stat stat_buf;

        int x, status;
        char parent[PATH_MAX];
        char *strptr;

	u_int64_t head_chunk_size;
	u_int64_t contents_chunk_size;

	u_int32_t h_device;
	u_int32_t h_inode;
	u_int64_t h_permissions;
	u_int32_t h_hardlinks;
	u_int32_t h_uid;
	u_int32_t h_gid;
	u_int32_t h_devtype;
	u_int64_t h_filesize;
	u_int64_t h_blocksize;
	u_int64_t h_nujmblocks;
	u_int64_t h_atime;
	u_int64_t h_mtime;
	u_int64_t h_ctime;

	char link_dest[PATH_MAX + NAME_MAX];


	/* Nothing to move? */
	if(path == NULL)
	    return(-1);

	/* Path must exist, use lstat to check if its a link. */
	if(lstat(path, &stat_buf))
	    return(-1);



	/* Cannot recycle dirs or devices. */
	if(S_ISDIR(stat_buf.st_mode))
	{
	    fprintf(stderr, "Cannot add directories to recycle bin.\n");
	    return(-4);
	}
	else if(S_ISCHR(stat_buf.st_mode))
	{
            fprintf(stderr, "Cannot add character devices to recycle bin.\n");
	    return(-4);
	}
	else if(S_ISBLK(stat_buf.st_mode))
	{
	    fprintf(stderr, "Cannot add block devices to recycle bin.\n");
	    return(-4);
	}
	else if(S_ISFIFO(stat_buf.st_mode))
	{
	    fprintf(stderr, "Cannot add FIFO pipes to recycle bin.\n");
	    return(-4);
	}
	else if(S_ISSOCK(stat_buf.st_mode))
	{
	    fprintf(stderr, "Cannot add sockets to recycle bin.\n");
	    return(-4);
	}


	/* Check recycle bin path, create as needed. */
	if(SETUPRECYCLEBIN(recbin))
	    return(-1);


	/* ************************************************************** */

	/* Get parent path. */
        strptr = GetParentDir(recbin);
        strncpy(
            parent,
            (strptr == NULL) ? RECYCLE_GLOBAL_DIR : strptr,
            PATH_MAX
        );
        parent[PATH_MAX - 1] = '\0';


	/* Get target path (recycle bin). */
	strncpy(tar_path, recbin, PATH_MAX + NAME_MAX);
	tar_path[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Get source path (object). */
	strncpy(src_path, path, PATH_MAX + NAME_MAX);
        src_path[PATH_MAX + NAME_MAX - 1] = '\0';


	/* Cannot backup recycle bin itself. */
	if(!strcmp(src_path, tar_path))
	{
	    fprintf(stderr,
		"Cannot backup recycle bin itself.\n"
	    );
	    return(-5);
	}


	/* ************************************************************* */
	/* Check source object (real object). */
	lstat(src_path, &stat_buf);

	/* Backup link. */
	if(S_ISLNK(stat_buf.st_mode))
	{
            /* Open target file (recycle bin). */
            tar_fp = fopen(tar_path, "ab");
            if(tar_fp == NULL)
                return(-1);   

            /* Get source object stats (real object). */
            lstat(src_path, &stat_buf);

	    status = readlink(src_path, link_dest, PATH_MAX + NAME_MAX);
	    if(status < 0)
	    {
		fclose(tar_fp); tar_fp = NULL;
		return(-1);
	    }
	    if(status <= (PATH_MAX + NAME_MAX))
	        link_dest[status] = '\0';
	    link_dest[PATH_MAX + NAME_MAX - 1] = '\0';

            /* Calculate source file stats. */
            head_chunk_size =
                88 +                            /* Fields. */
                (strlen(src_path) + 1) +        /* Name + null byte. */
                (strlen(link_dest) + 1)		/* Link dest + null byte. */
            ;

            contents_chunk_size = 8;		/* No contencts. */

            h_device = stat_buf.st_dev;
            h_inode = stat_buf.st_ino;
            h_permissions = stat_buf.st_mode;
            h_hardlinks = stat_buf.st_nlink;
            h_uid = stat_buf.st_uid;
            h_gid = stat_buf.st_gid;
            h_devtype = stat_buf.st_rdev;
            h_filesize = stat_buf.st_size;
            h_blocksize = stat_buf.st_blksize;
            h_nujmblocks = stat_buf.st_blocks;
            h_atime = stat_buf.st_atime;
            h_mtime = stat_buf.st_mtime;
            h_ctime = stat_buf.st_ctime;

            /* Write header chunk. */
            fwrite(&head_chunk_size, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_device, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_inode, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_permissions, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_hardlinks, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_uid, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_gid, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_devtype, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_filesize, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_blocksize, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_nujmblocks, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_atime, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_mtime, sizeof(u_int64_t), 1, tar_fp);  
            fwrite(&h_ctime, sizeof(u_int64_t), 1, tar_fp);
 
            fwrite(src_path, sizeof(char), strlen(src_path), tar_fp);
            fputc('\0', tar_fp);

	    fwrite(link_dest, sizeof(char), strlen(link_dest), tar_fp);
            fputc('\0', tar_fp);        /* For the link destination (none). */


            /* Write contents. */
            fwrite(&contents_chunk_size, sizeof(u_int64_t), 1, tar_fp);
  

            /* Close files. */
            fclose(tar_fp); tar_fp = NULL;
	}
	/* Backup file. */
	else
	{
	    /* Open target file (recycle bin). */
	    tar_fp = fopen(tar_path, "ab");
	    if(tar_fp == NULL)
	        return(-1);

	    /* Open source file (real object). */
	    src_fp = fopen(src_path, "rb");
	    if(src_fp == NULL)
	    {
	        fclose(tar_fp); tar_fp = NULL;
	        return(-1);
	    }

	    /* Get stats of source file (real object). */
	    fstat(fileno(src_fp), &stat_buf);


	    /* Calculate source file stats. */
	    head_chunk_size =
	        88 +				/* Fields. */
	        (strlen(src_path) + 1) +	/* Name + null byte. */
	        0 + 1				/* Link dest + null byte. */
	    ;

	    contents_chunk_size = 8 + stat_buf.st_size;

	    h_device = stat_buf.st_dev;
            h_inode = stat_buf.st_ino;
            h_permissions = stat_buf.st_mode;
            h_hardlinks = stat_buf.st_nlink;
            h_uid = stat_buf.st_uid;
            h_gid = stat_buf.st_gid;
            h_devtype = stat_buf.st_rdev;
            h_filesize = stat_buf.st_size;
            h_blocksize = stat_buf.st_blksize;
            h_nujmblocks = stat_buf.st_blocks;
            h_atime = stat_buf.st_atime;
            h_mtime = stat_buf.st_mtime;
            h_ctime = stat_buf.st_ctime;

	    /* Write header chunk. */
	    fwrite(&head_chunk_size, sizeof(u_int64_t), 1, tar_fp);
	    fwrite(&h_device, sizeof(u_int32_t), 1, tar_fp);
	    fwrite(&h_inode, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_permissions, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_hardlinks, sizeof(u_int32_t), 1, tar_fp);
	    fwrite(&h_uid, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_gid, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_devtype, sizeof(u_int32_t), 1, tar_fp);
            fwrite(&h_filesize, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_blocksize, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_nujmblocks, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_atime, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_mtime, sizeof(u_int64_t), 1, tar_fp);
            fwrite(&h_ctime, sizeof(u_int64_t), 1, tar_fp);

	    fwrite(src_path, sizeof(char), strlen(src_path), tar_fp);
	    fputc('\0', tar_fp);

	    fputc('\0', tar_fp);	/* For the link destination (none). */


	    /* Write contents. */
	    fwrite(&contents_chunk_size, sizeof(u_int64_t), 1, tar_fp);

	    for(x = 0; x < (int)h_filesize; x++)
	    {
	        fputc(fgetc(src_fp), tar_fp);
	    }


	    /* Close files. */
	    fclose(tar_fp); tar_fp = NULL;
	    fclose(src_fp); src_fp = NULL;
	}



	return(0);
}



/*
 *	Permanently removes entry number givin by entry_num (zero relative)
 *	from the recycle bin.
 *
 *	WARNING: This function will permanently remove the entry
 *	from the recycle bin WITHOUT restoring it!  Make sure you
 *	restore if FIRST before calling this function!!!
 */
int RecycleRemoveObject(int entry_num, char *recbin)
{
        FILE *src_fp;
	FILE *tar_fp;

        char src_path[PATH_MAX + NAME_MAX];
        char tar_path[PATH_MAX + NAME_MAX];

	off_t src_filesize;
        off_t src_next_pos;
        off_t src_cur_pos;

        u_int64_t chunk_size;        
        struct stat stat_buf;
	int i, cur_entry, status;

	char removed_object = 0;


        /* Nothing to move? */
        if(entry_num < 0)
            return(-1);
 
        /* Recycle bin must exist. */
        if(access(recbin, F_OK) == -1)
            return(-2);


	/* Get target path name (new recycle bin). */
	strncpy(tar_path, recbin, PATH_MAX + NAME_MAX);
	tar_path[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Get source path name (original recycle bin). */
	strncpy(src_path, recbin, PATH_MAX + NAME_MAX);
        src_path[PATH_MAX + NAME_MAX - 1] = '\0';
	strncat(src_path, ".old",
	    PATH_MAX + NAME_MAX -
	    strlen(src_path) - 1
	);
	src_path[PATH_MAX + NAME_MAX - 1] = '\0';

        /* Cannot have same names for both old and new recycle bins. */
        if(!strcmp(src_path, tar_path))
        {
            fprintf(stderr,
      "Unable to create new recycle bin, possible name length too long.\n"
            );
            return(-5);
        }


	/* Rename target to source. */
	status = rename(tar_path, src_path);
	if(status != 0)
	{
	    fprintf(stderr,
		"Unable to rename recycle bin for processing.\n"
	    );
	    return(-1);
	}

	/* *********************************************************** */

	src_fp = fopen(src_path, "rb");
	if(src_fp == NULL)
	{
	    fprintf(stderr,
		"%s: Cannot open.\n", src_path
	    );
	    return(-1);
	}
	fstat(fileno(src_fp), &stat_buf);
	src_filesize = stat_buf.st_size;

        tar_fp = fopen(tar_path, "wb");
        if(tar_fp == NULL)
        {
            fprintf(stderr,
                "%s: Cannot create.\n", tar_path
            );
	    fclose(src_fp); src_fp = NULL;
            return(-1);
        }


	cur_entry = 0;
	src_cur_pos = 0;
	src_next_pos = 0;

	/*
	 *	Begin copying over source recbin to target recbin,
	 *	skipping the entry specified in entry_num.
	 */
	while(src_next_pos < src_filesize)
	{
	    /* Do not copy over entry_num. */
	    if(cur_entry != entry_num)
	    {
		/* Copy over this entry. */

		/* Get header/attributes chunk size. */
		fseek(src_fp, src_cur_pos, SEEK_SET);
		fread(&chunk_size, sizeof(u_int64_t), 1, src_fp);
		src_next_pos += chunk_size;

		/* Copy over header. */
		fseek(src_fp, src_cur_pos, SEEK_SET);
		for(i = 0; i < (int)chunk_size; i++)
		{
		    fputc(fgetc(src_fp), tar_fp);
		}
		src_cur_pos = src_next_pos;


		/* Get contents chunk size. */
		fseek(src_fp, src_cur_pos, SEEK_SET);
                fread(&chunk_size, sizeof(u_int64_t), 1, src_fp);
                src_next_pos += chunk_size;

                /* Copy over contents. */ 
                fseek(src_fp, src_cur_pos, SEEK_SET);
                for(i = 0; i < (int)chunk_size; i++)
                {
                    fputc(fgetc(src_fp), tar_fp);
                }
                src_cur_pos = src_next_pos;
	    }
	    else
	    {
		/* Skip this entry. */

                /* Get header/attributes chunk size. */
                fseek(src_fp, src_cur_pos, SEEK_SET);
                fread(&chunk_size, sizeof(u_int64_t), 1, src_fp);
                src_next_pos += chunk_size;
		src_cur_pos = src_next_pos;

                /* Get contents chunk size. */
                fseek(src_fp, src_cur_pos, SEEK_SET);
                fread(&chunk_size, sizeof(u_int64_t), 1, src_fp);
                src_next_pos += chunk_size;
                src_cur_pos = src_next_pos;


		/* Mark that we have `removed' the object. */
		removed_object = 1;
	    }


	    cur_entry++;
	}



	/* Close the files. */
	fclose(src_fp); src_fp = NULL;
	fclose(tar_fp); tar_fp = NULL;

	/* Remove the original recycle bin. */
	RB_DELETE_REALOBJECT(src_path);


	return(0);
}


/*
 *	Attempts to recover object givin at entry_num in
 *	recbin to the real object path.
 *
 *	The real object path must not exist, if path is NULL then
 *	the original path of entry_num in the recbin will be used
 *	as the path provided that it does not exist.
 */
int RecycleRecoverObject(int entry_num, char *recbin, char *path)
{
	FILE *fp;
	int entry_count;
	struct stat stat_buf;
	u_int64_t chunksize;
	off_t filesize;

	long next_pos, cur_pos;

	FILE *tar_fp;
	char recover_path[PATH_MAX + NAME_MAX];
	char link_dest[PATH_MAX + NAME_MAX];
	int i, status;

        u_int32_t h_device;
        u_int32_t h_inode;
        u_int64_t h_permissions;
        u_int32_t h_hardlinks;
        u_int32_t h_uid;
        u_int32_t h_gid;
        u_int32_t h_devtype;
        u_int64_t h_filesize;
        u_int64_t h_blocksize;
        u_int64_t h_numblocks;
        u_int64_t h_atime;
        u_int64_t h_mtime;
        u_int64_t h_ctime;
        

	/* entry_num must be valid. */
	if(entry_num < 0)
	    return(-1);


	fp = fopen(recbin, "rb");
	if(fp == NULL)
	{
	    fprintf(stderr,
		"%s: Cannot open.\n",
                recbin
	    );
	    return(-1);
	}

        fstat(fileno(fp), &stat_buf);
        filesize = stat_buf.st_size;

	status = -1;
        next_pos = 0;
        cur_pos = 0;
	entry_count = 0;

        while(next_pos < filesize)
	{
	    if(entry_count == entry_num)
	    {
		/* This is the entry we want to recover. */

		/* Read header/attributes chunk. */
                fseek(fp, cur_pos, SEEK_SET);
                fread(&chunksize, sizeof(u_int64_t), 1, fp);
                next_pos += chunksize;

		/* Get name. */
		if(path == NULL)
		{
		    fseek(fp, cur_pos + 88, SEEK_SET);
		    fread(recover_path, sizeof(char), PATH_MAX + NAME_MAX, fp);
		    recover_path[PATH_MAX + NAME_MAX - 1] = '\0';
		    path = recover_path;
		}
		if(access(path, F_OK) == 0)
		{
		    fprintf(stderr, "%s: Already exists.\n", path);
		    status = -3;
		    break;
		}
		/* Attributes. */
		fseek(fp, cur_pos + 8, SEEK_SET);
		fread(&h_device, sizeof(u_int32_t), 1, fp);
		fread(&h_inode, sizeof(u_int32_t), 1, fp);
		fread(&h_permissions, sizeof(u_int64_t), 1, fp);
                fread(&h_hardlinks, sizeof(u_int32_t), 1, fp);
                fread(&h_uid, sizeof(u_int32_t), 1, fp);
                fread(&h_gid, sizeof(u_int32_t), 1, fp);
                fread(&h_devtype, sizeof(u_int32_t), 1, fp);
                fread(&h_filesize, sizeof(u_int64_t), 1, fp);
                fread(&h_blocksize, sizeof(u_int64_t), 1, fp);
                fread(&h_numblocks, sizeof(u_int64_t), 1, fp);
                fread(&h_atime, sizeof(u_int64_t), 1, fp);
                fread(&h_mtime, sizeof(u_int64_t), 1, fp);
                fread(&h_ctime, sizeof(u_int64_t), 1, fp);

		/* Link destination. */
		fseek(fp, cur_pos + 88, SEEK_SET);
		i = fgetc(fp);
		while((i != '\0') && (i != EOF))
		    i = fgetc(fp);
		fread(link_dest, sizeof(char), PATH_MAX + NAME_MAX, fp);
		link_dest[PATH_MAX + NAME_MAX - 1] = '\0';

		/* Recover link. */
		if(S_ISLNK(h_permissions))
		{
		    status = symlink(link_dest, path);
		}
		/* Recover file. */
		else if(S_ISREG(h_permissions))
		{
		    tar_fp = fopen(path, "wb");
		    if(tar_fp == NULL)
		    {
		        fprintf(stderr, "%s: Cannot create.\n", path);
		        status = -1;
		        break;
		    }

		    /* Set current position to start of content chunk. */
                    cur_pos = next_pos;

		    /* Recover contents. */
		    fseek(fp, cur_pos, SEEK_SET);
		    fread(&chunksize, sizeof(u_int64_t), 1, fp);
                    next_pos += chunksize;

		    for(i = 0;
                        i < (int)((long)chunksize - sizeof(u_int64_t));
                        i++
		    )
		    {
		        fputc(fgetc(fp), tar_fp);
		    }

		    fclose(tar_fp); tar_fp = NULL;

		    /* Set the file's attributes. */
		    chmod(path, h_permissions);

		    status = 0;		/* Success. */
		}
                /* Recover directory. */
                else if(S_ISDIR(h_permissions))
                {
		    status = mkdir(path, h_permissions);
		}
		/* Fifo pipe. */
		else if(S_ISFIFO(h_permissions))
		{
                    status = mkfifo(path, h_permissions);
		}


		/* Finished with recovery. */
                cur_pos = next_pos;

		break;
	    }
	    else
	    {
		/* Skip entry. */

                /* Read header/attributes chunk. */
                fseek(fp, cur_pos, SEEK_SET);
                fread(&chunksize, sizeof(u_int64_t), 1, fp);
                next_pos += chunksize;
		cur_pos = next_pos;

                /* Read contencts chunk. */
                fseek(fp, cur_pos, SEEK_SET);
                fread(&chunksize, sizeof(u_int64_t), 1, fp);
                next_pos += chunksize;
		cur_pos = next_pos;
	    }

	    entry_count++;
	}


        fclose(fp); fp = NULL;

	return(status);
}




/* ********************************************************************* */

/*
 *	Removes (purges) an object matching string from the recycle bin
 *	specified as recbin.
 *
 *	WARNING: The object matched in string will be permanently erased.
 */
void RecycleDoPurgeObject(char *string, char *recbin)
{
	int entry_num;


	/* Get object number. */
	entry_num = GETENTRYNUMBYNAME(string, recbin);
	if(entry_num < 0)
	    return;

	/* Remove entry_num from recbin. */
        RecycleRemoveObject(entry_num, recbin);


	return;
}


/*
 *	Adds an object specified as path (must be an absolute
 *	path) to the recycle bin specified as recbin.
 *
 *	If the original real object is NOT a directory then it is
 *	unlink()'ed.
 *
 *	If the original real object is a directory then the directory
 *	is checked for additional contents. If it has contents then
 *	those contents will be deleted by a for() loop to this function
 *	to unlink() each one of those files. Should one of those files be
 *	a dir then it will be recursed into as well.
 */
int RecycleDoRecycleObject(const char *path, const char *recbin)
{
	char *strptr;
	char tmp_path[PATH_MAX + NAME_MAX];
	struct stat stat_buf;


	if(path == NULL)
	    return(-1);

	strncpy(tmp_path, path, PATH_MAX + NAME_MAX);
	StringStripSpaces(tmp_path);

	/* Path must be absolute. */
	if(!ISPATHABSOLUTE(tmp_path))
	{
	    fprintf(stderr,
 "RecycleDoRecycleObject(): %s: Not absolute path.\n",
		tmp_path
	    );
	    return(-2);
	}

        /* Path must exist, use lstat to check if its a link. */
        if(lstat(tmp_path, &stat_buf))
            return(-1);

	/* Directory? */
	if(S_ISDIR(stat_buf.st_mode))
	{
	    char **strv, **strv_base;
	    char tmp_path2[PATH_MAX + NAME_MAX];

	    /* It is a directory, so get listing of its contents and
	     * recycle each one. Then recycle the dir. Note that this
	     * will cause recusion if one or more of the contents is
	     * a dir.
	     */
	    strv_base = GetDirEntNames(tmp_path);
	    strv = strv_base;
	    if(strv != NULL)
	    {
		while(*strv != NULL)
		{
		    if(!strcmp(*strv, ".") ||
                       !strcmp(*strv, "..")
		    )
		    {
			free(*strv);
			strv++;
			continue;
		    }

		    strptr = PrefixPaths(tmp_path, *strv);
		    if(strptr != NULL)
		    {
			strncpy(tmp_path2, strptr, PATH_MAX + NAME_MAX);
			tmp_path2[PATH_MAX + NAME_MAX - 1] = '\0';

			RecycleDoRecycleObject(tmp_path2, recbin);
		    }

		    free(*strv);
		    strv++;
		}

		free(strv_base);
	    }

	    /* Any objects still left in the directory? */
	    if(NUMDIRCONTENTS(tmp_path) > 0)
	    {
		/* Still items left in the directory. */
		char text[PATH_MAX + NAME_MAX + 256];

		sprintf(text,
"Recursive recycle halted abnormally, unrecycleable objects\n\
encountered in directory:\n\n\
    %s\n",
		    tmp_path
		);
                printdw(&err_dw, text);
	    }
	    else
	    {
		/* Directory is empty, safe to remove it. */
                rmdir(tmp_path);
            }
	}
	/* Character or block device? */
	else if(S_ISCHR(stat_buf.st_mode) ||
                S_ISBLK(stat_buf.st_mode)
        )
	{
            printdw(&err_dw,
"Cannot recycle character or block devices.\n\n"
            );
	}
	/* Recycle type FIFO pipe. */
	else if(S_ISFIFO(stat_buf.st_mode))
	{
	    /* Do not add FIFO pipes to recycle bin, just delete it. */

	    RB_DELETE_REALOBJECT(tmp_path);
	}
        /* Recycle sockets. */
        else if(S_ISSOCK(stat_buf.st_mode))
        {
	    /* Do not add sockets to recycle bin, just delete it. */

            RB_DELETE_REALOBJECT(tmp_path);
        }
	/* Files, links and all others. */
	else
	{
	    /* Add object specified in path to recycle bin. */
	    if(RecycleAddObject(tmp_path, recbin))
	    {
		/* Could not recycle it, so report error. */
	        return(-1);
	    }
	    else
	    {
		/* Erase the real object. */
		RB_DELETE_REALOBJECT(tmp_path);
	    }
	}


	return(0);
}



/*
 *	Procedure to restore an object matching string from
 *	the recycle bin specified as recbin to the target
 *	destination path as a real object.
 *
 *	If the destination path is pointed to NULL then the
 *	object will be restored to the path recored in the
 *	recycle bin (original path where the object was when
 *	it was initially recycled).
 *
 *	If the restore is successful, the recycled object
 *	will be purged.
 */
int RecycleDoRestoreObject(char *string, char *recbin, char *path)
{
	int entry_num, status;

	/* Get entry number. */
	entry_num = GETENTRYNUMBYNAME(string, recbin);
	if(entry_num < 0)
	    return(-1);

	/* Recover. */
	status = RecycleRecoverObject(entry_num, recbin, path);
	if(status != 0)
	{
	    fprintf(stderr,
		"%s: Unable to recover object from %s.\n",
		string, recbin
	    );
	    return(-1);
	}

	/* Now remove from recbin. */
	RecycleRemoveObject(entry_num, recbin);



	return(0);
}



/*
 *	Returns an allocated array of object structures
 *	containing the attributes and names of all objects in the
 *	recycle bin recbin.
 *
 *	The returned array of strings need to be free'ed afterwards!
 */
object_struct **RecycleGetObjectListing(char *recbin, int *nobjs)
{
	int i;
	char name[PATH_MAX + NAME_MAX];
        char link_dest[PATH_MAX + NAME_MAX];

	FILE *fp;
	off_t next_pos;
	off_t filesize;
	off_t cur_pos;

	u_int64_t chunk_size;
   
        u_int32_t h_device;
        u_int32_t h_inode;
        u_int64_t h_permissions;
        u_int32_t h_hardlinks;
        u_int32_t h_uid;
        u_int32_t h_gid;
        u_int32_t h_devtype;
        u_int64_t h_filesize;
        u_int64_t h_blocksize;
        u_int64_t h_numblocks;
        u_int64_t h_atime;
        u_int64_t h_mtime;
        u_int64_t h_ctime;

	struct stat stat_buf;

	object_struct **object;


	/* Reset objects. */
	object = NULL;

	/* Error checks. */
	if(nobjs == NULL)
	    return(object);

	*nobjs = 0;

	/* Path to recbin must exist. */
	if(recbin == NULL)
	    return(object);
	if(access(recbin, F_OK) == -1)
	    return(object);

	/* Open file. */
	fp = fopen(recbin, "rb");
	if(fp == NULL)
	{
	    fprintf(stderr, "%s: Cannot open.\n", recbin);
	    return(object);
	}
	fstat(fileno(fp), &stat_buf);


	/* *********************************************************** */

	next_pos = 0;
	cur_pos = 0;
	filesize = stat_buf.st_size;

	while(next_pos < filesize)
	{
	    /* Read HEADER/ATTRIBUTES chuck. */
	    fread(&chunk_size, sizeof(u_int64_t), 1, fp);
	    next_pos += chunk_size;

            /* Get attributes. */
            fread(&h_device, sizeof(u_int32_t), 1, fp);
            fread(&h_inode, sizeof(u_int32_t), 1, fp);
            fread(&h_permissions, sizeof(u_int64_t), 1, fp);
            fread(&h_hardlinks, sizeof(u_int32_t), 1, fp);
            fread(&h_uid, sizeof(u_int32_t), 1, fp);
            fread(&h_gid, sizeof(u_int32_t), 1, fp);
            fread(&h_devtype, sizeof(u_int32_t), 1, fp);
            fread(&h_filesize, sizeof(u_int64_t), 1, fp);
            fread(&h_blocksize, sizeof(u_int64_t), 1, fp);
            fread(&h_numblocks, sizeof(u_int64_t), 1, fp);
            fread(&h_atime, sizeof(u_int64_t), 1, fp); 
            fread(&h_mtime, sizeof(u_int64_t), 1, fp);
            fread(&h_ctime, sizeof(u_int64_t), 1, fp);

	    /* Get name. */
	    fseek(fp, cur_pos + 88, SEEK_SET);
	    fread(name, sizeof(char), PATH_MAX + NAME_MAX, fp);
	    name[PATH_MAX + NAME_MAX - 1] = '\0';

	    /* Get link_dest. */
	    fseek(fp, cur_pos + 88, SEEK_SET);
	    i = fgetc(fp);
	    while((i != '\0') && (i != EOF))
		i = fgetc(fp);
	    fread(link_dest, sizeof(char), PATH_MAX + NAME_MAX, fp);
	    link_dest[PATH_MAX + NAME_MAX - 1] = '\0';

	    /* Allocate new object structure. */
	    *nobjs += 1;
	    object = (object_struct **)realloc(object,
		*nobjs * sizeof(object_struct *));
	    if(object == NULL)
	    {
		*nobjs = 0;
		fclose(fp); fp = NULL;
		return(object);
	    }
	    i = (*nobjs - 1);
	    object[i] = (object_struct *)calloc(1, sizeof(object_struct));
	    if(object[i] == NULL)
	    {
		(*nobjs) = ((*nobjs) - 1);
                fclose(fp);
		fp = NULL;
                return(object);
	    }

	    object[i]->name = (char *)StringCopyAlloc(name);
	    object[i]->link_dest = (char *)StringCopyAlloc(link_dest);

	    object[i]->inode = h_inode;
            object[i]->mode = h_permissions;
            object[i]->nlink = h_hardlinks;
            object[i]->uid = h_uid;
            object[i]->gid = h_gid;
            object[i]->rdevice = h_devtype;
            object[i]->size = h_filesize;
            object[i]->last_accessed = h_atime;
            object[i]->last_modified = h_mtime;
            object[i]->create_time = h_ctime;



	    /* Seek next chunk. */
	    fseek(fp, next_pos, SEEK_SET);
	    cur_pos = next_pos;



	    /* Read CONTENTS chunk. */
            fread(&chunk_size, sizeof(u_int64_t), 1, fp);
            next_pos += chunk_size;

	    /* Seek next header. */
            fseek(fp, next_pos, SEEK_SET);
	    cur_pos = next_pos;
	}


	/* Close the file. */
	fclose(fp); fp = NULL;

	return(object);
}
