/*
                        Program Spool File Handling

	Functions:

	void SFRemoveProgStdout()
	void SFRemoveList()
	void SFRemoveAll()

	int SFDoStartupWarn()
	int SFDoShutdownWarn()

	int SFViewerProgStdout()
	int SFShowOutputOnDialog(dialog_win_struct *dw)

	---

	Spool files contain redirected output (stdout and/or stderr)
	of its (Endeavour's) child processes.

	These functions handle the removing of and reading data from.

 */

#include "endeavour.h"



/*
 *	Procedure to reset the file referanced in
 *	fname.prog_output.
 */
void SFRemoveProgStdout()
{
	struct stat stat_buf;


	/*   Check if the program output file exists, follow links
	 *   as needed.
	 */
	if(stat(fname.prog_output, &stat_buf))
	    return;

	/*   Truncate it, no point in removing it since it'll just
	 *   be created again.
	 */
	truncate(fname.prog_output, 0);


	return;
}


/*
 *      Procedure to reset the file referanced in
 *      fname.list_output.
 */
void SFRemoveList()
{
        struct stat stat_buf;
   
        
        /*   Check if the program output file exists, follow links
         *   as needed.
         */
        if(stat(fname.list_output, &stat_buf))  
            return;
        
        /*   Truncate it, no point in removing it since it'll just
         *   be created again.
         */
        truncate(fname.list_output, 0);

 
        return;
}



/*
 *	Procedure to remove all spool files.
 */
void SFRemoveAll()
{
	unlink(fname.prog_output);
	unlink(fname.list_output);

	return;
}


/* *********************************************************************** */

/*
 *	Procedure to check any existing spool files on start up and
 *	perform nessasary checks and print warnings.
 *
 *	This function should be called once at startup.
 */
int SFDoStartupWarn()
{
	char stringa[PATH_MAX + NAME_MAX + 128];
	char tmp_path[PATH_MAX + NAME_MAX];
	char **strv;
	int strc;
	struct stat stat_buf;


	/* Check for previous spool files. */
	if(!stat(dname.prog_home, &stat_buf))
	{
	    strv = GetDirEntNames(dname.prog_home);

	    if(strv != NULL)
	    {
		for(strc = 0; strv[strc] != NULL; strc++)
	        {
		    strncpy(tmp_path, fname.prog_output_ori,
			PATH_MAX + NAME_MAX);
                    StripAbsolutePath(tmp_path);
		    if(strpfx(strv[strc], tmp_path))
		    {
			/* Print warning. */
			sprintf(stringa,
"Warning: A program output file exists, possibly created\n\
by a previous process of this program. You should manually\n\
remove all files with a name starting as\n\
`%s' in the %s directory.\n",
			    tmp_path,
			    dname.prog_home
			);
			printdw(&err_dw, stringa);
		    }

                    strncpy(tmp_path, fname.list_output_ori,
                        PATH_MAX + NAME_MAX);
		    StripAbsolutePath(tmp_path);
                    if(strpfx(strv[strc], tmp_path))
                    {
                        /* Print warning. */
                        sprintf(stringa,
"Warning: A list output file exists, possibly created\n\
by a previous process of this program. You should manually\n\
remove all files with a name starting as\n\
`%s' in the %s directory.\n",
                            tmp_path,
                            dname.prog_home
                        );
                        printdw(&err_dw, stringa); 
                    }

		    free(strv[strc]);
		}
		free(strv);
		strc = 0;
	    }
	}




	return(0);
}



/*
 *	Procedure to check spool files on shut down and perform
 *	nessasary checks and print warnings.
 *
 *	This function should be called on normal shutdown.
 */
int SFDoShutdownWarn()
{





	return(0);
}


/* *********************************************************************** */

/*
 *	Creates an internal viewer to view the program output spool file.
 */
int SFViewerProgStdout()
{
	int i;

        if(access(fname.prog_output, F_OK) == 0)
	{
	    i = IntViewerAllocate();
	    if(IntViewerIsAllocated(i))
	    {
                ViewerInit(pri_fviewer[i], 0, 0, 0, 0);

                ViewerLoadFile(pri_fviewer[i], fname.prog_output);
                ViewerMap(pri_fviewer[i]);
	    }
	}


	return(0);
}


/*
 *	Shows (as needed) the output from the output file on the
 *	dialog widget.
 */
int SFShowOutputOnDialog(dialog_win_struct *dw)
{
	FILE *fp;
	struct stat stat_buf;
	off_t filesize;
	char *buf;


	/* Check if display is connected. */
	if(!IDC() ||
           (dw == NULL)
	)
	    return(-1);


	/* Get stats and make sure output file exists, report no error. */
	if(stat(fname.prog_output, &stat_buf) == -1)
	    return(0);

	/* Get file size and allocate buffer. */
	filesize = stat_buf.st_size;
	if(filesize == 0)
	{
	    /* No contents. */
	    SFRemoveProgStdout();
	    return(0);
	}
	if(filesize > 256)	/* Too big to fit on dialog? */
	    return(0);

        buf = (char *)calloc(1, (filesize + 1) * sizeof(char));
        if(buf == NULL)
            return(-1);


	/* *********************************************************** */

	/* Open file. */
	fp = fopen(fname.prog_output, "rb");
	if(fp == NULL)
	    return(-1);

	/* Read buffer. */
	fread(buf, sizeof(char), filesize, fp);

	/* Close file. */
	fclose(fp); fp = NULL;


	/* *********************************************************** */

	/* Map dialog. */
	printdw(dw, buf);

	/* Free buffer, no longer needed. */
	free(buf); buf = NULL;


	return(0);
}
